import json
from base64 import b64encode
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import (
    AnymailAPIError,
    AnymailRequestsAPIError,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import attach_inline_image, attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("mailjet")
@override_settings(
    EMAIL_BACKEND="anymail.backends.mailjet.EmailBackend",
    ANYMAIL={
        "MAILJET_API_KEY": "API KEY HERE",
        "MAILJET_SECRET_KEY": "SECRET KEY HERE",
    },
)
class MailjetBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b"""{
        "Messages": [{
            "Status": "success",
            "To": [{
                "Email": "to@example.com",
                "MessageUUID": "cb927469-36fd-4c02-bce4-0d199929a207",
                "MessageID": 70650219165027410,
                "MessageHref": "https://api.mailjet.com/v3/message/70650219165027410"
            }]
        }]
    }"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("mailjet")
class MailjetBackendStandardEmailTests(MailjetBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("/v3.1/send")

        auth = self.get_api_call_auth()
        self.assertEqual(auth, ("API KEY HERE", "SECRET KEY HERE"))

        data = self.get_api_call_json()
        self.assertEqual(len(data["Messages"]), 1)
        message = data["Messages"][0]
        self.assertEqual(data["Globals"]["Subject"], "Subject here")
        self.assertEqual(data["Globals"]["TextPart"], "Here is the message.")
        self.assertEqual(data["Globals"]["From"], {"Email": "from@sender.example.com"})
        self.assertEqual(message["To"], [{"Email": "to@example.com"}])

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ['"Recipient, #1" <to1@example.com>', "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["Messages"]), 1)
        message = data["Messages"][0]
        self.assertEqual(
            data["Globals"]["From"], {"Email": "from@example.com", "Name": "From Name"}
        )
        self.assertEqual(
            message["To"],
            [
                {"Email": "to1@example.com", "Name": "Recipient, #1"},
                {"Email": "to2@example.com"},
            ],
        )
        self.assertEqual(
            data["Globals"]["Cc"],
            [
                {"Email": "cc1@example.com", "Name": "Carbon Copy"},
                {"Email": "cc2@example.com"},
            ],
        )
        self.assertEqual(
            data["Globals"]["Bcc"],
            [
                {"Email": "bcc1@example.com", "Name": "Blind Copy"},
                {"Email": "bcc2@example.com"},
            ],
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={"Reply-To": "another@example.com", "X-MyHeader": "my value"},
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["Messages"]), 1)
        message = data["Messages"][0]
        self.assertEqual(data["Globals"]["Subject"], "Subject")
        self.assertEqual(data["Globals"]["TextPart"], "Body goes here")
        self.assertEqual(data["Globals"]["From"], {"Email": "from@example.com"})
        self.assertEqual(
            message["To"],
            [
                {"Email": "to1@example.com"},
                {"Email": "to2@example.com", "Name": "Also To"},
            ],
        )
        self.assertEqual(
            data["Globals"]["Cc"],
            [
                {"Email": "cc1@example.com"},
                {"Email": "cc2@example.com", "Name": "Also CC"},
            ],
        )
        self.assertEqual(
            data["Globals"]["Bcc"],
            [
                {"Email": "bcc1@example.com"},
                {"Email": "bcc2@example.com", "Name": "Also BCC"},
            ],
        )
        # Reply-To should be moved to own param:
        self.assertEqual(data["Globals"]["Headers"], {"X-MyHeader": "my value"})
        self.assertEqual(data["Globals"]["ReplyTo"], {"Email": "another@example.com"})

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()

        data = self.get_api_call_json()
        self.assertEqual(len(data["Messages"]), 1)
        self.assertEqual(data["Globals"]["TextPart"], text_content)
        self.assertEqual(data["Globals"]["HTMLPart"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("Attachments", data["Globals"])

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()

        data = self.get_api_call_json()
        self.assertNotIn("TextPart", data["Globals"])
        self.assertEqual(data["Globals"]["HTMLPart"], html_content)

    def test_extra_headers(self):
        self.message.extra_headers = {"X-Custom": "string", "X-Num": 123}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["Globals"]["Headers"],
            {
                "X-Custom": "string",
                "X-Num": 123,
            },
        )

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    @override_settings(ANYMAIL_IGNORE_UNSUPPORTED_FEATURES=True)
    def test_reply_to(self):
        # Mailjet only allows single reply-to. Verify correct handling for that
        # when ignoring unsupported features.
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
            headers={"X-Other": "Keep"},
        )
        email.send()
        data = self.get_api_call_json()
        # only the first reply_to:
        self.assertEqual(data["Globals"]["ReplyTo"], {"Email": "reply@example.com"})
        # don't lose other headers:
        self.assertEqual(data["Globals"]["Headers"], {"X-Other": "Keep"})

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["Globals"]["Attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["Filename"], "test.txt")
        self.assertEqual(attachments[0]["ContentType"], "text/plain")
        self.assertEqual(
            decode_att(attachments[0]["Base64Content"]).decode("ascii"), text_content
        )
        self.assertNotIn("ContentID", attachments[0])

        # inferred from filename:
        self.assertEqual(attachments[1]["ContentType"], "image/png")
        self.assertEqual(attachments[1]["Filename"], "test.png")
        self.assertEqual(decode_att(attachments[1]["Base64Content"]), png_content)
        # make sure image not treated as inline:
        self.assertNotIn("ContentID", attachments[1])

        self.assertEqual(attachments[2]["ContentType"], "application/pdf")
        self.assertEqual(attachments[2]["Filename"], "attachment")
        self.assertEqual(decode_att(attachments[2]["Base64Content"]), pdf_content)
        self.assertNotIn("ContentID", attachments[2])

        self.assertNotIn("InlinedAttachments", data["Globals"])

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["Globals"]["Attachments"],
            [
                {
                    "Filename": "Une pièce jointe.html",
                    "ContentType": "text/html",
                    "Base64Content": b64encode("<p>\u2019</p>".encode("utf-8")).decode(
                        "ascii"
                    ),
                }
            ],
        )

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        cid2 = attach_inline_image(self.message, image_data)
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["HTMLPart"], html_content)

        attachments = data["Globals"]["InlinedAttachments"]
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0]["Filename"], image_filename)
        self.assertEqual(attachments[0]["ContentID"], cid)
        self.assertEqual(attachments[0]["ContentType"], "image/png")
        self.assertEqual(decode_att(attachments[0]["Base64Content"]), image_data)
        # Mailjet requires a filename for all attachments, so make sure it's not empty:
        self.assertEqual(attachments[1]["Filename"], "attachment")
        self.assertEqual(attachments[1]["ContentID"], cid2)
        self.assertEqual(attachments[1]["ContentType"], "image/png")
        self.assertEqual(decode_att(attachments[1]["Base64Content"]), image_data)

        self.assertNotIn("Attachments", data["Globals"])

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file:
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly:
        image = MIMEImage(image_data)
        self.message.attach(image)

        image_data_b64 = b64encode(image_data).decode("ascii")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["Globals"]["Attachments"],
            [
                {
                    "Filename": image_filename,  # the named one
                    "ContentType": "image/png",
                    "Base64Content": image_data_b64,
                },
                {
                    "Filename": "attachment",  # the unnamed one
                    "ContentType": "image/png",
                    "Base64Content": image_data_b64,
                },
            ],
        )

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty to, cc, bcc, and reply_to shouldn't generate empty fields"""
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("Cc", data["Globals"])
        self.assertNotIn("Bcc", data["Globals"])
        self.assertNotIn("ReplyTo", data["Globals"])

    def test_empty_to_list(self):
        # Mailjet v3.1 doesn't support cc-only or bcc-only messages
        self.message.to = []
        self.message.cc = ["cc@example.com"]
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "messages without any `to` recipients"
        ):
            self.message.send()

    def test_api_failure(self):
        self.set_mock_response(status_code=500)
        with self.assertRaisesMessage(AnymailAPIError, "Mailjet API response 500"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=500)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        # JSON error response - global error:
        error_response = json.dumps(
            {
                "ErrorIdentifier": "06df1144-c6f3-4ca7-8885-7ec5d4344113",
                "ErrorCode": "mj-0002",
                "ErrorMessage": "Helpful explanation from Mailjet.",
                "StatusCode": 400,
            }
        ).encode("utf-8")
        self.set_mock_response(status_code=400, raw=error_response)
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful explanation from Mailjet"
        ):
            self.message.send()

        # Non-JSON error response:
        self.set_mock_response(status_code=500, raw=b"Ack! Bad proxy!")
        with self.assertRaisesMessage(AnymailAPIError, "Ack! Bad proxy!"):
            self.message.send()

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()


@tag("mailjet")
class MailjetBackendAnymailFeatureTests(MailjetBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "bounce-handler@bounces.example.com"
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["Globals"]["Sender"], {"Email": "bounce-handler@bounces.example.com"}
        )

    def test_metadata(self):
        # Mailjet expects the payload to be a single string
        # https://dev.mailjet.com/guides/#tagging-email-messages
        self.message.metadata = {"user_id": "12345", "items": 6}
        self.message.send()
        data = self.get_api_call_json()
        self.assertJSONEqual(
            data["Globals"]["EventPayload"], {"user_id": "12345", "items": 6}
        )

    def test_send_at(self):
        self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "send_at"):
            self.message.send()

    def test_tags(self):
        self.message.tags = ["receipt"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["CustomCampaign"], "receipt")

        self.message.tags = ["receipt", "repeat-user"]
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "multiple tags"):
            self.message.send()

    def test_track_opens(self):
        self.message.track_opens = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["TrackOpens"], "enabled")

    def test_track_clicks(self):
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["TrackClicks"], "enabled")

        self.message.track_clicks = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["TrackClicks"], "disabled")

    def test_template(self):
        # template_id can be str or int (but must be numeric ID-not the template's name)
        self.message.template_id = "1234567"
        self.message.merge_global_data = {"name": "Alice", "group": "Developers"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["TemplateID"], 1234567)  # must be integer
        # TemplateLanguage required to use variables:
        self.assertEqual(data["Globals"]["TemplateLanguage"], True)
        self.assertEqual(
            data["Globals"]["Variables"], {"name": "Alice", "group": "Developers"}
        )

    def test_template_populate_from_sender(self):
        # v3.1 API allows omitting From param to use template's sender
        self.message.template_id = "1234567"
        # must set from_email to None after constructing EmailMessage:
        self.message.from_email = None
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("From", data["Globals"])  # use template's sender as From

    def test_merge_data(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},
        }
        self.message.merge_global_data = {
            "group": "Default Group",
            "global": "Global value",
        }
        self.message.send()
        data = self.get_api_call_json()
        messages = data["Messages"]
        # with merge_data, each 'to' gets separate message:
        self.assertEqual(len(messages), 2)

        self.assertEqual(messages[0]["To"], [{"Email": "alice@example.com"}])
        self.assertEqual(
            messages[1]["To"], [{"Email": "bob@example.com", "Name": "Bob"}]
        )

        # global merge_data is sent in Globals
        self.assertEqual(
            data["Globals"]["Variables"],
            {"group": "Default Group", "global": "Global value"},
        )

        # per-recipient merge_data is sent in Messages
        # (and Mailjet will merge with Globals)
        self.assertEqual(
            messages[0]["Variables"], {"name": "Alice", "group": "Developers"}
        )
        self.assertEqual(messages[1]["Variables"], {"name": "Bob"})

    def test_merge_metadata(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},
        }
        self.message.metadata = {"notification_batch": "zx912"}
        self.message.send()

        data = self.get_api_call_json()
        messages = data["Messages"]
        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[0]["To"][0]["Email"], "alice@example.com")
        # metadata and merge_metadata[recipient] are combined:
        self.assertJSONEqual(
            messages[0]["EventPayload"],
            {"order_id": 123, "tier": "premium", "notification_batch": "zx912"},
        )
        self.assertEqual(messages[1]["To"][0]["Email"], "bob@example.com")
        self.assertJSONEqual(
            messages[1]["EventPayload"],
            {"order_id": 678, "notification_batch": "zx912"},
        )

    def test_merge_headers(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "List-Unsubscribe": "<https://example.com/a/>",
            },
            "bob@example.com": {
                "List-Unsubscribe": "<https://example.com/b/>",
            },
        }
        self.message.send()

        data = self.get_api_call_json()
        messages = data["Messages"]
        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[0]["To"][0]["Email"], "alice@example.com")
        self.assertEqual(
            messages[0]["Headers"],
            {"List-Unsubscribe": "<https://example.com/a/>"},
        )
        self.assertEqual(messages[1]["To"][0]["Email"], "bob@example.com")
        self.assertEqual(
            messages[1]["Headers"],
            {"List-Unsubscribe": "<https://example.com/b/>"},
        )

        # non-merge headers still in globals:
        self.assertEqual(
            data["Globals"]["Headers"],
            {
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
        )

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("CustomCampaign", data["Globals"])
        self.assertNotIn("EventPayload", data["Globals"])
        self.assertNotIn("HTMLPart", data["Globals"])
        self.assertNotIn("TemplateID", data["Globals"])
        self.assertNotIn("TemplateLanguage", data["Globals"])
        self.assertNotIn("Variables", data["Globals"])
        self.assertNotIn("TrackOpens", data["Globals"])
        self.assertNotIn("TrackClicks", data["Globals"])

    def test_esp_extra(self):
        # Anymail deep merges Mailjet esp_extra into the v3.1 Send API payload.
        # Most options you'd want to override are in Globals, though a few are
        # at the root. Note that it's *not* possible to merge into Messages
        # (though you could completely replace it).
        self.message.esp_extra = {
            "Globals": {
                "TemplateErrorDeliver": True,
                "TemplateErrorReporting": "bugs@example.com",
            },
            "SandboxMode": True,
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Globals"]["TemplateErrorDeliver"], True)
        self.assertEqual(data["Globals"]["TemplateErrorReporting"], "bugs@example.com")
        self.assertIs(data["SandboxMode"], True)
        # Make sure the backend params are also still there
        self.assertEqual(data["Globals"]["Subject"], "Subject")

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        response_content = json.dumps(
            {
                "Messages": [
                    {
                        "Status": "success",
                        "To": [
                            {
                                "Email": "to1@example.com",
                                "MessageUUID": "cb927469-36fd-4c02-bce4-0d199929a207",
                                "MessageID": 12345678901234500,
                                "MessageHref": "https://api.mailjet.com/v3/message"
                                "/12345678901234500",
                            }
                        ],
                    }
                ]
            }
        ).encode("utf-8")
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject", "Message", "from@example.com", ["to1@example.com"]
        )
        sent = msg.send()

        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"sent"})
        self.assertEqual(msg.anymail_status.message_id, "12345678901234500")
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "sent"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            "12345678901234500",
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_mixed_status(self):
        """The status should include an entry for each recipient"""
        # Mailjet's v3.1 API will partially fail a batch send, allowing valid emails
        # to go out. The API response doesn't identify the failed email addresses;
        # make sure we represent them correctly in the anymail_status.
        response_data = {
            "Messages": [
                {
                    "Status": "success",
                    "CustomID": "",
                    "To": [
                        {
                            "Email": "to-good@example.com",
                            "MessageUUID": "556e896a-e041-4836-bb35-8bb75ee308c5",
                            "MessageID": 12345678901234500,
                            "MessageHref": "https://api.mailjet.com/v3/REST"
                            "/message/12345678901234500",
                        }
                    ],
                    "Cc": [],
                    "Bcc": [],
                },
                {
                    "Errors": [
                        {
                            "ErrorIdentifier": "f480a5a2-0334-4e08"
                            "-b2b7-f372ce5669e0",
                            "ErrorCode": "mj-0013",
                            "StatusCode": 400,
                            "ErrorMessage": '"invalid@123.4" is an invalid'
                            " email address.",
                            "ErrorRelatedTo": ["To[0].Email"],
                        }
                    ],
                    "Status": "error",
                },
            ]
        }
        # Mailjet uses 400 for partial success:
        self.set_mock_response(json_data=response_data, status_code=400)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to-good@example.com", "invalid@123.4"],
        )
        sent = msg.send()

        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"sent", "failed"})
        self.assertEqual(
            msg.anymail_status.recipients["to-good@example.com"].status, "sent"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to-good@example.com"].message_id,
            "12345678901234500",
        )
        self.assertEqual(
            msg.anymail_status.recipients["invalid@123.4"].status, "failed"
        )
        self.assertEqual(
            msg.anymail_status.recipients["invalid@123.4"].message_id, None
        )
        self.assertEqual(msg.anymail_status.message_id, {"12345678901234500", None})
        self.assertEqual(msg.anymail_status.esp_response.json(), response_data)

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    def test_non_json_error(self):
        """429 (and other?) errors don't have a json payload"""
        self.set_mock_response(
            status_code=429, raw=b"Rate limit exceeded", content_type="text/html"
        )
        with self.assertRaisesRegex(
            AnymailRequestsAPIError,
            r"^Mailjet API response 429 .*Rate limit exceeded.*",
        ) as cm:
            self.message.send()
        self.assertNotIn("Invalid JSON", str(cm.exception))

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but a non-JSON API response, should raise an API exception
        """
        mock_response = self.set_mock_response(
            status_code=200, raw=b"yikes, this isn't a real response"
        )
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, mock_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.tags = [Decimal("19.99")]  # yeah, don't do this
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_json())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Mailjet", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")

    def test_merge_data_null_values(self):
        # Mailjet doesn't accept None (null) as a merge value;
        # returns "HTTP/1.1 500 Cannot convert data from Null value"
        self.message.merge_global_data = {"Some": None}
        self.set_mock_response(
            status_code=500, reason="Cannot convert data from Null value", raw=None
        )
        with self.assertRaisesMessage(
            AnymailAPIError, "Cannot convert data from Null value"
        ):
            self.message.send()


@tag("mailjet")
class MailjetBackendSessionSharingTestCase(
    SessionSharingTestCases, MailjetBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("mailjet")
@override_settings(EMAIL_BACKEND="anymail.backends.mailjet.EmailBackend")
class MailjetBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_api_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bMAILJET_API_KEY\b")

    @override_settings(ANYMAIL={"MAILJET_API_KEY": "dummy"})
    def test_missing_secret_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bMAILJET_SECRET_KEY\b")



# Tests for the anymail/utils.py module
# (not to be confused with utilities for testing found in tests/utils.py)
import base64
import copy
import pickle
from email.mime.image import MIMEImage
from email.mime.text import MIMEText

from django.http import QueryDict
from django.test import RequestFactory, SimpleTestCase, override_settings
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy

from anymail.exceptions import AnymailInvalidAddress, _LazyError
from anymail.utils import (
    UNSET,
    Attachment,
    CaseInsensitiveCasePreservingDict,
    EmailAddress,
    concat_lists,
    force_non_lazy,
    force_non_lazy_dict,
    force_non_lazy_list,
    get_request_basic_auth,
    get_request_uri,
    is_lazy,
    last,
    merge_dicts_deep,
    merge_dicts_one_level,
    merge_dicts_shallow,
    parse_address_list,
    parse_rfc2822date,
    parse_single_address,
    querydict_getfirst,
    update_deep,
)


class ParseAddressListTests(SimpleTestCase):
    """Test utils.parse_address_list"""

    def test_simple_email(self):
        parsed_list = parse_address_list(["test@example.com"])
        self.assertEqual(len(parsed_list), 1)
        parsed = parsed_list[0]
        self.assertIsInstance(parsed, EmailAddress)
        self.assertEqual(parsed.addr_spec, "test@example.com")
        self.assertEqual(parsed.display_name, "")
        self.assertEqual(parsed.address, "test@example.com")
        self.assertEqual(parsed.username, "test")
        self.assertEqual(parsed.domain, "example.com")

    def test_display_name(self):
        parsed_list = parse_address_list(['"Display Name, Inc." <test@example.com>'])
        self.assertEqual(len(parsed_list), 1)
        parsed = parsed_list[0]
        self.assertEqual(parsed.addr_spec, "test@example.com")
        self.assertEqual(parsed.display_name, "Display Name, Inc.")
        self.assertEqual(parsed.address, '"Display Name, Inc." <test@example.com>')
        self.assertEqual(parsed.username, "test")
        self.assertEqual(parsed.domain, "example.com")

    def test_obsolete_display_name(self):
        # you can get away without the quotes if there are no commas or parens
        # (but it's not recommended)
        parsed_list = parse_address_list(["Display Name <test@example.com>"])
        self.assertEqual(len(parsed_list), 1)
        parsed = parsed_list[0]
        self.assertEqual(parsed.addr_spec, "test@example.com")
        self.assertEqual(parsed.display_name, "Display Name")
        self.assertEqual(parsed.address, "Display Name <test@example.com>")

    def test_unicode_display_name(self):
        parsed_list = parse_address_list(
            ['"Unicode \N{HEAVY BLACK HEART}" <test@example.com>']
        )
        self.assertEqual(len(parsed_list), 1)
        parsed = parsed_list[0]
        self.assertEqual(parsed.addr_spec, "test@example.com")
        self.assertEqual(parsed.display_name, "Unicode \N{HEAVY BLACK HEART}")
        # formatted display-name automatically shifts
        # to quoted-printable/base64 for non-ascii chars:
        self.assertEqual(
            parsed.address, "=?utf-8?b?VW5pY29kZSDinaQ=?= <test@example.com>"
        )

    def test_invalid_display_name(self):
        with self.assertRaisesMessage(
            AnymailInvalidAddress, "Invalid email address 'webmaster'"
        ):
            parse_address_list(["webmaster"])

        with self.assertRaisesMessage(
            AnymailInvalidAddress, "Maybe missing quotes around a display-name?"
        ):
            # this parses as multiple email addresses, because of the comma:
            parse_address_list(["Display Name, Inc. <test@example.com>"])

    def test_idn(self):
        parsed_list = parse_address_list(["idn@\N{ENVELOPE}.example.com"])
        self.assertEqual(len(parsed_list), 1)
        parsed = parsed_list[0]
        self.assertEqual(parsed.addr_spec, "idn@\N{ENVELOPE}.example.com")
        # punycode-encoded domain:
        self.assertEqual(parsed.address, "idn@xn--4bi.example.com")
        self.assertEqual(parsed.username, "idn")
        self.assertEqual(parsed.domain, "\N{ENVELOPE}.example.com")

    def test_none_address(self):
        # used for, e.g., telling Mandrill to use template default from_email
        self.assertEqual(parse_address_list([None]), [])
        self.assertEqual(parse_address_list(None), [])

    def test_empty_address(self):
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list([""])

    def test_whitespace_only_address(self):
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list([" "])

    def test_invalid_address(self):
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list(["localonly"])
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list(["localonly@"])
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list(["@domainonly"])
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list(["<localonly@>"])
        with self.assertRaises(AnymailInvalidAddress):
            parse_address_list(["<@domainonly>"])

    def test_email_list(self):
        parsed_list = parse_address_list(["first@example.com", "second@example.com"])
        self.assertEqual(len(parsed_list), 2)
        self.assertEqual(parsed_list[0].addr_spec, "first@example.com")
        self.assertEqual(parsed_list[1].addr_spec, "second@example.com")

    def test_multiple_emails(self):
        # Django's EmailMessage allows multiple, comma-separated emails
        # in a single recipient string. (It passes them along to the backend intact.)
        # (Depending on this behavior is not recommended.)
        parsed_list = parse_address_list(["first@example.com, second@example.com"])
        self.assertEqual(len(parsed_list), 2)
        self.assertEqual(parsed_list[0].addr_spec, "first@example.com")
        self.assertEqual(parsed_list[1].addr_spec, "second@example.com")

    def test_invalid_in_list(self):
        # Make sure it's not just concatenating list items...
        # the bare "Display Name" below should *not* get merged with
        # the email in the second item
        with self.assertRaisesMessage(AnymailInvalidAddress, "Display Name"):
            parse_address_list(['"Display Name"', "<valid@example.com>"])

    def test_invalid_with_unicode(self):
        with self.assertRaisesMessage(
            AnymailInvalidAddress, "Invalid email address '\N{ENVELOPE}'"
        ):
            parse_address_list(["\N{ENVELOPE}"])

    def test_single_string(self):
        # bare strings are used by the from_email parsing in BasePayload
        parsed_list = parse_address_list("one@example.com")
        self.assertEqual(len(parsed_list), 1)
        self.assertEqual(parsed_list[0].addr_spec, "one@example.com")

    def test_lazy_strings(self):
        parsed_list = parse_address_list(
            [gettext_lazy('"Example, Inc." <one@example.com>')]
        )
        self.assertEqual(len(parsed_list), 1)
        self.assertEqual(parsed_list[0].display_name, "Example, Inc.")
        self.assertEqual(parsed_list[0].addr_spec, "one@example.com")

        parsed_list = parse_address_list(gettext_lazy("one@example.com"))
        self.assertEqual(len(parsed_list), 1)
        self.assertEqual(parsed_list[0].display_name, "")
        self.assertEqual(parsed_list[0].addr_spec, "one@example.com")

    def test_parse_one(self):
        parsed = parse_single_address("one@example.com")
        self.assertEqual(parsed.address, "one@example.com")

        with self.assertRaisesMessage(
            AnymailInvalidAddress, "Only one email address is allowed; found 2"
        ):
            parse_single_address("one@example.com, two@example.com")

        with self.assertRaisesMessage(AnymailInvalidAddress, "Invalid email address"):
            parse_single_address(" ")

    def test_no_newlines(self):
        # (Parsing shouldn't even be able to even generate these cases,
        # but in case anyone constructs an EmailAddress directly...)
        for name, addr in [
            ("Potential\nInjection", "addr@example.com"),
            ("Potential\rInjection", "addr@example.com"),
            ("Name", "potential\ninjection@example.com"),
            ("Name", "potential\rinjection@example.com"),
        ]:
            with self.subTest(name=name, addr=addr):
                with self.assertRaisesMessage(ValueError, "cannot contain newlines"):
                    _ = EmailAddress(name, addr)

    def test_email_address_repr(self):
        self.assertEqual(
            "EmailAddress('Name', 'addr@example.com')",
            repr(EmailAddress("Name", "addr@example.com")),
        )


class NormalizedAttachmentTests(SimpleTestCase):
    """Test utils.Attachment"""

    # (Several basic tests could be added here)

    def test_content_disposition_attachment(self):
        image = MIMEImage(b";-)", "x-emoticon")
        image["Content-Disposition"] = 'attachment; filename="emoticon.txt"'
        att = Attachment(image, "ascii")
        self.assertEqual(att.name, "emoticon.txt")
        self.assertEqual(att.content, b";-)")
        self.assertFalse(att.inline)
        self.assertIsNone(att.content_id)
        self.assertEqual(att.cid, "")
        self.assertEqual(
            repr(att), "Attachment<image/x-emoticon, len=3, name='emoticon.txt'>"
        )

    def test_content_disposition_inline(self):
        image = MIMEImage(b";-)", "x-emoticon")
        image["Content-Disposition"] = "inline"
        att = Attachment(image, "ascii")
        self.assertIsNone(att.name)
        self.assertEqual(att.content, b";-)")
        self.assertTrue(att.inline)  # even without the Content-ID
        self.assertIsNone(att.content_id)
        self.assertEqual(att.cid, "")
        self.assertEqual(
            repr(att), "Attachment<inline, image/x-emoticon, len=3, content_id=None>"
        )

        image["Content-ID"] = "<abc123@example.net>"
        att = Attachment(image, "ascii")
        self.assertEqual(att.content_id, "<abc123@example.net>")
        self.assertEqual(att.cid, "abc123@example.net")
        self.assertEqual(
            repr(att),
            "Attachment<inline, image/x-emoticon, len=3,"
            " content_id='<abc123@example.net>'>",
        )

    def test_content_id_implies_inline(self):
        """A MIME object with a Content-ID should be assumed to be inline"""
        image = MIMEImage(b";-)", "x-emoticon")
        image["Content-ID"] = "<abc123@example.net>"
        att = Attachment(image, "ascii")
        self.assertTrue(att.inline)
        self.assertEqual(att.content_id, "<abc123@example.net>")
        self.assertEqual(
            repr(att),
            "Attachment<inline, image/x-emoticon, len=3,"
            " content_id='<abc123@example.net>'>",
        )

        # ... but not if explicit Content-Disposition says otherwise
        image["Content-Disposition"] = "attachment"
        att = Attachment(image, "ascii")
        self.assertFalse(att.inline)
        self.assertIsNone(att.content_id)  # ignored for non-inline Attachment
        self.assertEqual(repr(att), "Attachment<image/x-emoticon, len=3>")

    def test_content_type(self):
        att = Attachment(MIMEText("text", "plain", "iso8859-1"), "ascii")
        self.assertEqual(att.mimetype, "text/plain")
        self.assertEqual(att.content_type, 'text/plain; charset="iso8859-1"')
        self.assertEqual(repr(att), "Attachment<text/plain, len=4>")


class LazyCoercionTests(SimpleTestCase):
    """Test utils.is_lazy and force_non_lazy*"""

    def test_is_lazy(self):
        self.assertTrue(is_lazy(gettext_lazy("lazy string is lazy")))

    def test_not_lazy(self):
        self.assertFalse(is_lazy("text not lazy"))
        self.assertFalse(is_lazy(b"bytes not lazy"))
        self.assertFalse(is_lazy(None))
        self.assertFalse(is_lazy({"dict": "not lazy"}))
        self.assertFalse(is_lazy(["list", "not lazy"]))
        self.assertFalse(is_lazy(object()))
        self.assertFalse(is_lazy([gettext_lazy("doesn't recurse")]))

    def test_force_lazy(self):
        result = force_non_lazy(gettext_lazy("text"))
        self.assertIsInstance(result, str)
        self.assertEqual(result, "text")

    def test_format_lazy(self):
        self.assertTrue(
            is_lazy(
                format_lazy(
                    "{0}{1}", gettext_lazy("concatenation"), gettext_lazy("is lazy")
                )
            )
        )
        result = force_non_lazy(
            format_lazy(
                "{first}/{second}",
                first=gettext_lazy("text"),
                second=gettext_lazy("format"),
            )
        )
        self.assertIsInstance(result, str)
        self.assertEqual(result, "text/format")

    def test_force_string(self):
        result = force_non_lazy("text")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "text")

    def test_force_bytes(self):
        result = force_non_lazy(b"bytes \xFE")
        self.assertIsInstance(result, bytes)
        self.assertEqual(result, b"bytes \xFE")

    def test_force_none(self):
        result = force_non_lazy(None)
        self.assertIsNone(result)

    def test_force_dict(self):
        result = force_non_lazy_dict(
            {"a": 1, "b": gettext_lazy("b"), "c": {"c1": gettext_lazy("c1")}}
        )
        self.assertEqual(result, {"a": 1, "b": "b", "c": {"c1": "c1"}})
        self.assertIsInstance(result["b"], str)
        self.assertIsInstance(result["c"]["c1"], str)

    def test_force_list(self):
        result = force_non_lazy_list([0, gettext_lazy("b"), "c"])
        self.assertEqual(result, [0, "b", "c"])  # coerced to list
        self.assertIsInstance(result[1], str)


class UpdateDeepTests(SimpleTestCase):
    """Test utils.update_deep"""

    def test_updates_recursively(self):
        first = {"a": {"a1": 1, "aa": {}}, "b": "B"}
        second = {"a": {"a2": 2, "aa": {"aa1": 11}}}
        result = update_deep(first, second)
        self.assertEqual(first, {"a": {"a1": 1, "a2": 2, "aa": {"aa1": 11}}, "b": "B"})
        # modifies first in place; doesn't return it (same as dict.update()):
        self.assertIsNone(result)

    def test_overwrites_sequences(self):
        """Only mappings are handled recursively; sequences are considered atomic"""
        first = {"a": [1, 2]}
        second = {"a": [3]}
        update_deep(first, second)
        self.assertEqual(first, {"a": [3]})

    def test_handles_non_dict_mappings(self):
        """Mapping types in general are supported"""
        from collections import OrderedDict, defaultdict

        first = OrderedDict(a=OrderedDict(a1=1), c={"c1": 1})
        second = defaultdict(None, a=dict(a2=2))
        update_deep(first, second)
        self.assertEqual(first, {"a": {"a1": 1, "a2": 2}, "c": {"c1": 1}})


@override_settings(ALLOWED_HOSTS=[".example.com"])
class RequestUtilsTests(SimpleTestCase):
    """Test utils.get_request_* helpers"""

    def setUp(self):
        self.request_factory = RequestFactory()
        super().setUp()

    @staticmethod
    def basic_auth(username, password):
        """
        Return HTTP_AUTHORIZATION header value for basic auth with username, password
        """
        credentials = base64.b64encode(
            "{}:{}".format(username, password).encode("utf-8")
        ).decode("utf-8")
        return "Basic {}".format(credentials)

    def test_get_request_basic_auth(self):
        # without auth:
        request = self.request_factory.post(
            "/path/to/?query", HTTP_HOST="www.example.com", HTTP_SCHEME="https"
        )
        self.assertIsNone(get_request_basic_auth(request))

        # with basic auth:
        request = self.request_factory.post(
            "/path/to/?query",
            HTTP_HOST="www.example.com",
            HTTP_AUTHORIZATION=self.basic_auth("user", "pass"),
        )
        self.assertEqual(get_request_basic_auth(request), "user:pass")

        # with some other auth
        request = self.request_factory.post(
            "/path/to/?query",
            HTTP_HOST="www.example.com",
            HTTP_AUTHORIZATION="Bearer abcde12345",
        )
        self.assertIsNone(get_request_basic_auth(request))

    def test_get_request_uri(self):
        # without auth:
        request = self.request_factory.post(
            "/path/to/?query", secure=True, HTTP_HOST="www.example.com"
        )
        self.assertEqual(
            get_request_uri(request), "https://www.example.com/path/to/?query"
        )

        # with basic auth:
        request = self.request_factory.post(
            "/path/to/?query",
            secure=True,
            HTTP_HOST="www.example.com",
            HTTP_AUTHORIZATION=self.basic_auth("user", "pass"),
        )
        self.assertEqual(
            get_request_uri(request), "https://user:pass@www.example.com/path/to/?query"
        )

    @override_settings(
        SECURE_PROXY_SSL_HEADER=("HTTP_X_FORWARDED_PROTO", "https"),
        USE_X_FORWARDED_HOST=True,
    )
    def test_get_request_uri_with_proxy(self):
        request = self.request_factory.post(
            "/path/to/?query",
            secure=False,
            HTTP_HOST="web1.internal",
            HTTP_X_FORWARDED_PROTO="https",
            HTTP_X_FORWARDED_HOST="secret.example.com:8989",
            HTTP_AUTHORIZATION=self.basic_auth("user", "pass"),
        )
        self.assertEqual(
            get_request_uri(request),
            "https://user:pass@secret.example.com:8989/path/to/?query",
        )


class QueryDictUtilsTests(SimpleTestCase):
    def test_querydict_getfirst(self):
        q = QueryDict("a=one&a=two&a=three")
        q.getfirst = querydict_getfirst.__get__(q)
        self.assertEqual(q.getfirst("a"), "one")

        # missing key exception:
        with self.assertRaisesMessage(KeyError, "not a key"):
            q.getfirst("not a key")

        # defaults:
        self.assertEqual(q.getfirst("not a key", "beta"), "beta")
        self.assertIsNone(q.getfirst("not a key", None))


class ParseRFC2822DateTests(SimpleTestCase):
    def test_with_timezones(self):
        dt = parse_rfc2822date("Tue, 24 Oct 2017 10:11:35 -0700")
        self.assertEqual(dt.isoformat(), "2017-10-24T10:11:35-07:00")
        self.assertIsNotNone(dt.utcoffset())  # aware

        dt = parse_rfc2822date("Tue, 24 Oct 2017 10:11:35 +0700")
        self.assertEqual(dt.isoformat(), "2017-10-24T10:11:35+07:00")
        self.assertIsNotNone(dt.utcoffset())  # aware

        dt = parse_rfc2822date("Tue, 24 Oct 2017 10:11:35 +0000")
        self.assertEqual(dt.isoformat(), "2017-10-24T10:11:35+00:00")
        self.assertIsNotNone(dt.tzinfo)  # aware

    def test_without_timezones(self):
        # "no timezone information":
        dt = parse_rfc2822date("Tue, 24 Oct 2017 10:11:35 -0000")
        self.assertEqual(dt.isoformat(), "2017-10-24T10:11:35")
        # naive (compare with +0000 version in previous test):
        self.assertIsNone(dt.tzinfo)

        dt = parse_rfc2822date("Tue, 24 Oct 2017 10:11:35")
        self.assertEqual(dt.isoformat(), "2017-10-24T10:11:35")
        self.assertIsNone(dt.tzinfo)  # naive

    def test_unparseable_dates(self):
        self.assertIsNone(parse_rfc2822date(""))
        self.assertIsNone(parse_rfc2822date("  "))
        self.assertIsNone(parse_rfc2822date("garbage"))
        self.assertIsNone(parse_rfc2822date("Tue, 24 Oct"))
        self.assertIsNone(parse_rfc2822date("Lug, 24 Nod 2017 10:11:35 +0000"))
        self.assertIsNone(parse_rfc2822date("Tue, 99 Oct 9999 99:99:99 +9999"))


class LazyErrorTests(SimpleTestCase):
    def test_attr(self):
        lazy = _LazyError(ValueError("lazy failure"))  # creating doesn't cause error
        lazy.some_prop = "foo"  # setattr doesn't cause error
        with self.assertRaisesMessage(ValueError, "lazy failure"):
            self.unused = lazy.anything  # getattr *does* cause error

    def test_call(self):
        lazy = _LazyError(ValueError("lazy failure"))  # creating doesn't cause error
        with self.assertRaisesMessage(ValueError, "lazy failure"):
            self.unused = lazy()  # call *does* cause error


class CaseInsensitiveCasePreservingDictTests(SimpleTestCase):
    def setUp(self):
        self.dict = CaseInsensitiveCasePreservingDict()
        self.dict["Accept"] = "application/text+xml"
        self.dict["accEPT"] = "application/json"

    def test_preserves_first_key(self):
        self.assertEqual(list(self.dict.keys()), ["Accept"])

    def test_copy(self):
        copy = self.dict.copy()
        self.assertIsNot(copy, self.dict)
        self.assertEqual(copy, self.dict)
        # Here's why the superclass CaseInsensitiveDict.copy is insufficient:
        self.assertIsInstance(copy, CaseInsensitiveCasePreservingDict)

    def test_get_item(self):
        self.assertEqual(self.dict["accept"], "application/json")
        self.assertEqual(self.dict["Accept"], "application/json")
        self.assertEqual(self.dict["accEPT"], "application/json")

    # The base CaseInsensitiveDict functionality is well-tested in Requests,
    # so we don't repeat it here.


class UnsetValueTests(SimpleTestCase):
    """Tests for the UNSET sentinel value"""

    def test_not_other_values(self):
        self.assertIsNot(UNSET, None)
        self.assertIsNot(UNSET, False)
        self.assertNotEqual(UNSET, 0)
        self.assertNotEqual(UNSET, "")

    def test_unset_survives_pickle(self):
        # Required for using AnymailMessage with django-mailer
        pickled = pickle.dumps(UNSET)
        self.assertIs(pickle.loads(pickled), UNSET)

    def test_unset_survives_copy(self):
        self.assertIs(copy.copy(UNSET), UNSET)
        self.assertIs(copy.deepcopy(UNSET), UNSET)

    def test_unset_has_useful_repr(self):
        # (something better than '<object object at ...>')
        self.assertIn("UNSET", repr(UNSET))

    def test_equality(self):
        # `is UNSET` is preferred to `== UNSET`, but both should work
        self.assertEqual(UNSET, UNSET)


class CombinerTests(SimpleTestCase):
    def test_concat_lists(self):
        for args, expected in [
            (([1, 2], [3, 4]), [1, 2, 3, 4]),
            # Does not flatten:
            (([1, [11, 12]], [2]), [1, [11, 12], 2]),
            # UNSET args ignored:
            ((UNSET, [1, 2], UNSET, [3, 4], UNSET), [1, 2, 3, 4]),
            # None clears previous:
            (([1, 2], None, [3, 4]), [3, 4]),
            # Works with other sequence-like types:
            (([1], (2, 3), {4}), [1, 2, 3, 4]),
            # Degenerate cases:
            ((), UNSET),
            ((UNSET,), UNSET),
            ((None,), UNSET),
            (([], None), UNSET),
        ]:
            with self.subTest(repr(args)):
                original_args = copy.deepcopy(args)
                merged = concat_lists(*args)
                self.assertEqual(merged, expected)
                # Verify args were not modified:
                self.assertEqual(args, original_args)

    def test_merge_dicts_shallow(self):
        for args, expected in [
            (({"a": 1}, {"b": 2}), {"a": 1, "b": 2}),
            (
                ({"a": 1, "b": 2}, {"a": 11, "c": 33}, {"c": 3}),
                {"a": 11, "b": 2, "c": 3},
            ),
            # shallow merge:
            (({"a": {"a1": 1}, "b": 2}, {"a": {"a2": 2}}), {"a": {"a2": 2}, "b": 2}),
            # UNSET args ignored:
            ((UNSET, {"a": 1}, UNSET, {"b": 2}, UNSET), {"a": 1, "b": 2}),
            # None clears previous:
            (({"a": 1}, None, {"b": 2}), {"b": 2}),
            # Degenerate cases:
            ((), UNSET),
            ((UNSET,), UNSET),
            ((None,), UNSET),
            (({}, None), UNSET),
        ]:
            with self.subTest(repr(args)):
                original_args = copy.deepcopy(args)
                merged = merge_dicts_shallow(*args)
                self.assertEqual(merged, expected)
                # Verify args were not modified:
                self.assertEqual(args, original_args)

    def test_merge_dicts_deep(self):
        for args, expected in [
            (({"a": 1}, {"b": 2}), {"a": 1, "b": 2}),
            (
                ({"a": 1, "b": 2}, {"a": 11, "c": 33}, {"c": 3}),
                {"a": 11, "b": 2, "c": 3},
            ),
            # deep merge:
            (
                (
                    {"a": {"a1": 1, "a3": {"a3a": 31}}},
                    {"a": {"a2": 2, "a3": {"a3b": 32}}},
                ),
                {"a": {"a1": 1, "a2": 2, "a3": {"a3a": 31, "a3b": 32}}},
            ),
            # UNSET (top-level) args ignored:
            ((UNSET, {"a": 1}, UNSET, {"b": 2}, UNSET), {"a": 1, "b": 2}),
            # None clears previous:
            (({"a": 1}, None, {"b": 2}), {"b": 2}),
            # Degenerate cases:
            ((), UNSET),
            ((UNSET,), UNSET),
            ((None,), UNSET),
            (({}, None), UNSET),
        ]:
            with self.subTest(repr(args)):
                original_args = copy.deepcopy(args)
                merged = merge_dicts_deep(*args)
                self.assertEqual(merged, expected)
                # Verify args were not modified:
                self.assertEqual(args, original_args)

    def test_merge_dicts_one_level(self):
        for args, expected in [
            # one-level merge:
            (
                (
                    {"a": {"a1": 1, "a3": {"a3a": 31}}},
                    {"a": {"a2": 2, "a3": {"a3b": 32}}},
                ),
                {"a": {"a1": 1, "a2": 2, "a3": {"a3b": 32}}},  # but not a3a
            ),
            # UNSET (top-level) args ignored:
            ((UNSET, {"a": {}}, UNSET, {"b": {}}, UNSET), {"a": {}, "b": {}}),
            # None clears previous:
            (({"a": {}}, None, {"b": {}}), {"b": {}}),
            # Degenerate cases:
            ((), UNSET),
            ((UNSET,), UNSET),
            ((None,), UNSET),
            (({}, None), UNSET),
        ]:
            with self.subTest(repr(args)):
                original_args = copy.deepcopy(args)
                merged = merge_dicts_one_level(*args)
                self.assertEqual(merged, expected)
                # Verify args were not modified:
                self.assertEqual(args, original_args)

    def test_last(self):
        for args, expected in [
            ((1, 2, 3), 3),
            # UNSET args ignored:
            ((UNSET, 1, UNSET, 2, UNSET), 2),
            # None clears previous:
            ((1, 2, None), UNSET),
            # Degenerate cases:
            ((), UNSET),
            ((UNSET,), UNSET),
            ((None,), UNSET),
        ]:
            with self.subTest(repr(args)):
                original_args = copy.deepcopy(args)
                merged = last(*args)
                self.assertEqual(merged, expected)
                # Verify args were not modified:
                self.assertEqual(args, original_args)



import quopri
from base64 import b64encode
from email.utils import collapse_rfc2231_value
from textwrap import dedent

from django.test import SimpleTestCase

from anymail.exceptions import AnymailDeprecationWarning
from anymail.inbound import AnymailInboundMessage

from .utils import SAMPLE_IMAGE_FILENAME, sample_email_path, sample_image_content

SAMPLE_IMAGE_CONTENT = sample_image_content()


class AnymailInboundMessageConstructionTests(SimpleTestCase):
    def test_construct_params(self):
        msg = AnymailInboundMessage.construct(
            from_email="from@example.com",
            to="to@example.com",
            cc="cc@example.com",
            subject="test subject",
        )
        self.assertEqual(msg["From"], "from@example.com")
        self.assertEqual(msg["To"], "to@example.com")
        self.assertEqual(msg["Cc"], "cc@example.com")
        self.assertEqual(msg["Subject"], "test subject")

        # ensures email.message.Message.__init__ ran:
        self.assertEqual(msg.defects, [])
        # ensures AnymailInboundMessage.__init__ ran:
        self.assertIsNone(msg.envelope_recipient)

    def test_construct_headers_from_mapping(self):
        msg = AnymailInboundMessage.construct(
            headers={"Reply-To": "reply@example.com", "X-Test": "anything"}
        )
        # headers are case-insensitive:
        self.assertEqual(msg["reply-to"], "reply@example.com")
        self.assertEqual(msg["X-TEST"], "anything")

    def test_construct_headers_from_pairs(self):
        # allows multiple instances of a header
        msg = AnymailInboundMessage.construct(
            headers=[
                ["Reply-To", "reply@example.com"],
                [
                    "Received",
                    "by 10.1.1.4 with SMTP id q4csp;"
                    " Sun, 22 Oct 2017 00:23:22 -0700 (PDT)",
                ],
                [
                    "Received",
                    "from mail.example.com (mail.example.com. [10.10.1.9])"
                    " by mx.example.com with SMTPS id 93s8iok for <to@example.com>;"
                    " Sun, 22 Oct 2017 00:23:21 -0700 (PDT)",
                ],
            ]
        )
        self.assertEqual(msg["Reply-To"], "reply@example.com")
        self.assertEqual(
            msg.get_all("Received"),
            [
                "by 10.1.1.4 with SMTP id q4csp; Sun, 22 Oct 2017 00:23:22 -0700 (PDT)",
                "from mail.example.com (mail.example.com. [10.10.1.9])"
                " by mx.example.com with SMTPS id 93s8iok for <to@example.com>;"
                " Sun, 22 Oct 2017 00:23:21 -0700 (PDT)",
            ],
        )

    def test_construct_headers_from_raw(self):
        # (note header "folding" in second Received header)
        msg = AnymailInboundMessage.construct(
            raw_headers=dedent(
                """\
                Reply-To: reply@example.com
                Subject: raw subject
                Content-Type: x-custom/custom
                Received: by 10.1.1.4 with SMTP id q4csp; Sun, 22 Oct 2017 00:23:22 -0700 (PDT)
                Received: from mail.example.com (mail.example.com. [10.10.1.9])
                 by mx.example.com with SMTPS id 93s8iok for <to@example.com>;
                 Sun, 22 Oct 2017 00:23:21 -0700 (PDT)
                """  # NOQA: E501
            ),
            subject="Explicit subject overrides raw",
        )
        self.assertEqual(msg["Reply-To"], "reply@example.com")
        self.assertEqual(
            msg.get_all("Received"),
            [
                # unfolding should have stripped newlines
                "by 10.1.1.4 with SMTP id q4csp; Sun, 22 Oct 2017 00:23:22 -0700 (PDT)",
                "from mail.example.com (mail.example.com. [10.10.1.9])"
                " by mx.example.com with SMTPS id 93s8iok for <to@example.com>;"
                " Sun, 22 Oct 2017 00:23:21 -0700 (PDT)",
            ],
        )
        self.assertEqual(msg.get_all("Subject"), ["Explicit subject overrides raw"])
        # Content-Type in raw header ignored:
        self.assertEqual(msg.get_all("Content-Type"), ["multipart/mixed"])

    def test_construct_bodies(self):
        # this verifies we construct the expected MIME structure;
        # see the `text` and `html` props (in the ConveniencePropTests below)
        # for an easier way to get to these fields (that works however constructed)
        msg = AnymailInboundMessage.construct(text="Plaintext body", html="HTML body")
        self.assertEqual(msg["Content-Type"], "multipart/mixed")
        self.assertEqual(len(msg.get_payload()), 1)

        related = msg.get_payload(0)
        self.assertEqual(related["Content-Type"], "multipart/related")
        self.assertEqual(len(related.get_payload()), 1)

        alternative = related.get_payload(0)
        self.assertEqual(alternative["Content-Type"], "multipart/alternative")
        self.assertEqual(len(alternative.get_payload()), 2)

        plaintext = alternative.get_payload(0)
        self.assertEqual(plaintext["Content-Type"], 'text/plain; charset="utf-8"')
        self.assertEqual(plaintext.get_content_text(), "Plaintext body")

        html = alternative.get_payload(1)
        self.assertEqual(html["Content-Type"], 'text/html; charset="utf-8"')
        self.assertEqual(html.get_content_text(), "HTML body")

    def test_construct_attachments(self):
        att1 = AnymailInboundMessage.construct_attachment(
            "text/csv",
            "One,Two\n1,2".encode("iso-8859-1"),
            charset="iso-8859-1",
            filename="test.csv",
        )

        att2 = AnymailInboundMessage.construct_attachment(
            "image/png",
            SAMPLE_IMAGE_CONTENT,
            filename=SAMPLE_IMAGE_FILENAME,
            content_id="abc123",
        )

        msg = AnymailInboundMessage.construct(attachments=[att1, att2])
        self.assertEqual(msg["Content-Type"], "multipart/mixed")
        self.assertEqual(len(msg.get_payload()), 2)  # bodies (related), att1

        att1_part = msg.get_payload(1)
        self.assertEqual(
            att1_part["Content-Type"], 'text/csv; name="test.csv"; charset="iso-8859-1"'
        )
        self.assertEqual(
            att1_part["Content-Disposition"], 'attachment; filename="test.csv"'
        )
        self.assertNotIn("Content-ID", att1_part)
        self.assertEqual(att1_part.get_content_text(), "One,Two\n1,2")

        related = msg.get_payload(0)
        # alternatives (with no bodies in this test); att2:
        self.assertEqual(len(related.get_payload()), 2)
        att2_part = related.get_payload(1)
        self.assertEqual(
            att2_part["Content-Type"], 'image/png; name="sample_image.png"'
        )
        self.assertEqual(
            att2_part["Content-Disposition"], 'inline; filename="sample_image.png"'
        )
        self.assertEqual(att2_part["Content-ID"], "<abc123>")
        self.assertEqual(att2_part.get_content_bytes(), SAMPLE_IMAGE_CONTENT)

    def test_construct_attachments_from_uploaded_files(self):
        from django.core.files.uploadedfile import SimpleUploadedFile

        file = SimpleUploadedFile(
            SAMPLE_IMAGE_FILENAME, SAMPLE_IMAGE_CONTENT, "image/png"
        )
        att = AnymailInboundMessage.construct_attachment_from_uploaded_file(
            file, content_id="abc123"
        )
        self.assertEqual(att["Content-Type"], 'image/png; name="sample_image.png"')
        self.assertEqual(
            att["Content-Disposition"], 'inline; filename="sample_image.png"'
        )
        self.assertEqual(att["Content-ID"], "<abc123>")
        self.assertEqual(att.get_content_bytes(), SAMPLE_IMAGE_CONTENT)

    def test_construct_attachments_from_base64_data(self):
        # This is a fairly common way for ESPs to provide attachment content to webhooks
        content = b64encode(SAMPLE_IMAGE_CONTENT)
        att = AnymailInboundMessage.construct_attachment(
            content_type="image/png", content=content, base64=True
        )
        self.assertEqual(att.get_content_bytes(), SAMPLE_IMAGE_CONTENT)

    def test_construct_attachment_unicode_filename(self):
        # Issue #197
        att = AnymailInboundMessage.construct_attachment(
            content_type="text/plain",
            content="Unicode ✓",
            charset="utf-8",
            base64=False,
            filename="Simulácia.txt",
            content_id="inline-id",
        )
        self.assertEqual(att.get_filename(), "Simulácia.txt")
        self.assertTrue(att.is_inline())
        self.assertEqual(att.get_content_text(), "Unicode ✓")

    def test_parse_raw_mime(self):
        # (we're not trying to exhaustively test email.parser MIME handling here;
        # just that AnymailInboundMessage.parse_raw_mime calls it correctly)
        raw = dedent(
            """\
            Content-Type: text/plain
            Subject: This is a test message

            This is a test body.
            """
        )
        msg = AnymailInboundMessage.parse_raw_mime(raw)
        self.assertEqual(msg["Subject"], "This is a test message")
        self.assertEqual(msg.get_content_text(), "This is a test body.\n")
        self.assertEqual(msg.defects, [])

    # (see test_attachment_as_uploaded_file below
    # for parsing basic attachment from raw mime)

    def test_parse_raw_mime_bytes(self):
        raw = (
            b"Content-Type: text/plain; charset=ISO-8859-3\r\n"
            b"Content-Transfer-Encoding: 8bit\r\n"
            b"Subject: Test bytes\r\n"
            b"\r\n"
            b"\xD8i estas retpo\xFEto.\r\n"
        )
        msg = AnymailInboundMessage.parse_raw_mime_bytes(raw)
        self.assertEqual(msg["Subject"], "Test bytes")
        self.assertEqual(msg.get_content_text(), "Ĝi estas retpoŝto.\r\n")
        self.assertEqual(msg.get_content_bytes(), b"\xD8i estas retpo\xFEto.\r\n")
        self.assertEqual(msg.defects, [])

    def test_parse_raw_mime_8bit_utf8(self):
        # In come cases, the message below ends up with
        # 'Content-Transfer-Encoding: 8bit', so needs to be parsed as bytes, not text
        # (see https://bugs.python.org/issue18271).
        # Message.as_string() returns str (text), not bytes.
        # (This might be a Django bug; plain old MIMEText avoids the problem by using
        # 'Content-Transfer-Encoding: base64', which parses fine as text or bytes.)
        # Either way, AnymailInboundMessage should try to sidestep the whole issue.
        # (This test might no longer be relevant--this might have been a corner case
        # in Django's SafeMIME code in Python 2.7 only.)
        raw = dedent(
            # django.core.mail.message.SafeMIMEText("Unicode ✓", "plain", "utf-8").as_string()
            """\
            Content-Type: text/plain; charset="utf-8"
            MIME-Version: 1.0
            Content-Transfer-Encoding: 8bit

            Unicode ✓"""
        )
        msg = AnymailInboundMessage.parse_raw_mime(raw)
        self.assertEqual(msg.text, "Unicode ✓")  # *not* "Unicode \\u2713"

    def test_parse_raw_mime_file_text(self):
        with open(sample_email_path(), mode="r") as fp:
            msg = AnymailInboundMessage.parse_raw_mime_file(fp)
        self.assertEqual(msg["Subject"], "Test email")
        self.assertEqual(msg.text, "Hi Bob, This is a message. Thanks!\n")
        self.assertEqual(
            msg.get_all("Received"),
            [  # this is the first line in the sample email file
                "by luna.mailgun.net with SMTP mgrt 8734663311733;"
                " Fri, 03 May 2013 18:26:27 +0000"
            ],
        )

    def test_parse_raw_mime_file_bytes(self):
        with open(sample_email_path(), mode="rb") as fp:
            msg = AnymailInboundMessage.parse_raw_mime_file(fp)
        self.assertEqual(msg["Subject"], "Test email")
        self.assertEqual(msg.text, "Hi Bob, This is a message. Thanks!\n")
        self.assertEqual(
            msg.get_all("Received"),
            [  # this is the first line in the sample email file
                "by luna.mailgun.net with SMTP mgrt 8734663311733;"
                " Fri, 03 May 2013 18:26:27 +0000"
            ],
        )


class AnymailInboundMessageConveniencePropTests(SimpleTestCase):
    # AnymailInboundMessage defines several properties to simplify reading
    # commonly-used items in an email.message.Message

    def test_address_props(self):
        msg = AnymailInboundMessage.construct(
            from_email='"Sender, Inc." <sender@example.com>',
            to="First To <to1@example.com>, to2@example.com",
            cc="First Cc <cc1@example.com>, cc2@example.com",
            bcc="bcc@example.com",
        )
        self.assertEqual(str(msg.from_email), '"Sender, Inc." <sender@example.com>')
        self.assertEqual(msg.from_email.addr_spec, "sender@example.com")
        self.assertEqual(msg.from_email.display_name, "Sender, Inc.")
        self.assertEqual(msg.from_email.username, "sender")
        self.assertEqual(msg.from_email.domain, "example.com")

        self.assertEqual(len(msg.to), 2)
        self.assertEqual(msg.to[0].addr_spec, "to1@example.com")
        self.assertEqual(msg.to[0].display_name, "First To")
        self.assertEqual(msg.to[1].addr_spec, "to2@example.com")
        self.assertEqual(msg.to[1].display_name, "")

        self.assertEqual(len(msg.cc), 2)
        self.assertEqual(msg.cc[0].address, "First Cc <cc1@example.com>")
        self.assertEqual(msg.cc[1].address, "cc2@example.com")

        self.assertEqual(len(msg.bcc), 1)
        self.assertEqual(msg.bcc[0].address, "bcc@example.com")

        # Default None/empty lists
        msg = AnymailInboundMessage()
        self.assertIsNone(msg.from_email)
        self.assertEqual(msg.to, [])
        self.assertEqual(msg.cc, [])

        # Empty strings
        msg = AnymailInboundMessage.construct(from_email="", to="", cc="")
        self.assertIsNone(msg.from_email)
        self.assertEqual(msg.to, [])
        self.assertEqual(msg.cc, [])

    def test_body_props(self):
        msg = AnymailInboundMessage.construct(text="Test plaintext", html="Test HTML")
        self.assertEqual(msg.text, "Test plaintext")
        self.assertEqual(msg.html, "Test HTML")

        # Make sure attachments don't confuse it
        att_text = AnymailInboundMessage.construct_attachment(
            "text/plain", "text attachment"
        )
        att_html = AnymailInboundMessage.construct_attachment(
            "text/html", "html attachment"
        )

        msg = AnymailInboundMessage.construct(
            text="Test plaintext", attachments=[att_text, att_html]
        )
        self.assertEqual(msg.text, "Test plaintext")
        self.assertIsNone(msg.html)  # no html body (the html attachment doesn't count)

        msg = AnymailInboundMessage.construct(
            html="Test HTML", attachments=[att_text, att_html]
        )
        # no plaintext body (the text attachment doesn't count):
        self.assertIsNone(msg.text)
        self.assertEqual(msg.html, "Test HTML")

        # Default None
        msg = AnymailInboundMessage()
        self.assertIsNone(msg.text)
        self.assertIsNone(msg.html)

    def test_body_props_charsets(self):
        text_8859_10 = "Detta är det vanliga innehållet".encode("ISO-8859-10")
        html_8859_8 = "<p>HTML זהו תוכן</p>".encode("ISO-8859-8")
        raw = dedent(
            """\
            MIME-Version: 1.0
            Subject: Charset test
            Content-Type: multipart/alternative; boundary="this_is_a_boundary"

            --this_is_a_boundary
            Content-Type: text/plain; charset=ISO-8859-10
            Content-Transfer-Encoding: QUOTED-PRINTABLE

            {text}
            --this_is_a_boundary
            Content-Type: text/html; charset=ISO-8859-8
            Content-Transfer-Encoding: QUOTED-PRINTABLE

            {html}
            --this_is_a_boundary--
            """
        ).format(
            text=quopri.encodestring(text_8859_10).decode("ASCII"),
            html=quopri.encodestring(html_8859_8).decode("ASCII"),
        )

        msg = AnymailInboundMessage.parse_raw_mime(raw)
        self.assertEqual(msg.defects, [])
        self.assertEqual(msg.text, "Detta är det vanliga innehållet")
        self.assertEqual(msg.html, "<p>HTML זהו תוכן</p>")

        self.assertEqual(msg.get_payload(0).get_content_bytes(), text_8859_10)
        self.assertEqual(
            msg.get_payload(0).get_content_text(), "Detta är det vanliga innehållet"
        )
        self.assertEqual(msg.get_payload(1).get_content_bytes(), html_8859_8)
        self.assertEqual(msg.get_payload(1).get_content_text(), "<p>HTML זהו תוכן</p>")

    def test_missing_or_invalid_charsets(self):
        """
        get_content_text has options for handling missing/invalid charset declarations
        """
        raw = dedent(
            """\
            Subject: Oops, missing charset declaration
            Content-Type: text/plain
            Content-Transfer-Encoding: quoted-printable

            Algunos programas de correo electr=f3nico est=e1n rotos
            """
        )
        msg = AnymailInboundMessage.parse_raw_mime(raw)
        self.assertEqual(msg.defects, [])

        # default is charset from Content-Type (or 'utf-8' if missing),
        # errors='replace'; .text uses defaults
        self.assertEqual(
            msg.get_content_text(),
            "Algunos programas de correo electr�nico est�n rotos\n",
        )
        self.assertEqual(
            msg.text, "Algunos programas de correo electr�nico est�n rotos\n"
        )

        # can give specific charset if you know headers are wrong/missing
        self.assertEqual(
            msg.get_content_text(charset="ISO-8859-1"),
            "Algunos programas de correo electrónico están rotos\n",
        )

        # can change error handling
        with self.assertRaises(UnicodeDecodeError):
            msg.get_content_text(errors="strict")
        self.assertEqual(
            msg.get_content_text(errors="ignore"),
            "Algunos programas de correo electrnico estn rotos\n",
        )

    def test_date_props(self):
        msg = AnymailInboundMessage.construct(
            headers={"Date": "Mon, 23 Oct 2017 17:50:55 -0700"}
        )
        self.assertEqual(msg.date.isoformat(), "2017-10-23T17:50:55-07:00")

        # Default None
        self.assertIsNone(AnymailInboundMessage().date)

    def test_attachments_prop(self):
        att = AnymailInboundMessage.construct_attachment(
            "image/png", SAMPLE_IMAGE_CONTENT, filename=SAMPLE_IMAGE_FILENAME
        )

        msg = AnymailInboundMessage.construct(attachments=[att])
        self.assertEqual(msg.attachments, [att])

        # Default empty list
        self.assertEqual(AnymailInboundMessage().attachments, [])

    def test_content_id_map_prop(self):
        att = AnymailInboundMessage.construct_attachment(
            "image/png",
            SAMPLE_IMAGE_CONTENT,
            filename=SAMPLE_IMAGE_FILENAME,
            content_id="abc123",
        )

        msg = AnymailInboundMessage.construct(attachments=[att])
        self.assertEqual(msg.content_id_map, {"abc123": att})

        with self.assertWarnsMessage(
            AnymailDeprecationWarning,
            "inline_attachments has been renamed to content_id_map and will be removed "
            "in the near future.",
        ):
            self.assertEqual(msg.inline_attachments, {"abc123": att})

        # Default empty dict
        self.assertEqual(AnymailInboundMessage().content_id_map, {})

    def test_inlines_prop(self):
        raw = dedent(
            """\
            MIME-Version: 1.0
            Subject: Message with inline parts
            Content-Type: multipart/mixed; boundary="boundary-orig"

            --boundary-orig
            Content-Type: text/html; charset="UTF-8"

            <img src="cid:abc123"> Here is a message!

            --boundary-orig
            Content-Type: image/png; name="sample_image.png"
            Content-Disposition: inline
            Content-ID: <abc123>
            Content-Transfer-Encoding: base64

            {image_content_base64}

            --boundary-orig
            Content-Type: image/png; name="sample_image_without_cid.png"
            Content-Disposition: inline
            Content-Transfer-Encoding: base64

            {image_content_base64}

            --boundary-orig--
            """
        ).format(image_content_base64=b64encode(SAMPLE_IMAGE_CONTENT).decode("ascii"))

        msg = AnymailInboundMessage.parse_raw_mime(raw)
        inlines = msg.inlines

        self.assertEqual(len(inlines), 2)

        self.assertEqual(inlines[0].get_content_type(), "image/png")
        self.assertEqual(inlines[0].as_uploaded_file().name, "sample_image.png")

        self.assertEqual(inlines[1].get_content_type(), "image/png")
        self.assertEqual(
            inlines[1].as_uploaded_file().name, "sample_image_without_cid.png"
        )

        self.assertEqual(len(msg.content_id_map.items()), 1)
        self.assertIn("abc123", msg.content_id_map)

    def test_attachment_as_uploaded_file(self):
        raw = dedent(
            """\
            MIME-Version: 1.0
            Subject: Attachment test
            Content-Type: multipart/mixed; boundary="this_is_a_boundary"

            --this_is_a_boundary
            Content-Type: text/plain; charset="UTF-8"

            The test sample image is attached below.

            --this_is_a_boundary
            Content-Type: image/png; name="sample_image.png"
            Content-Disposition: attachment; filename="sample_image.png"
            Content-Transfer-Encoding: base64

            iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlz
            AAALEgAACxIB0t1+/AAAABR0RVh0Q3JlYXRpb24gVGltZQAzLzEvMTNoZNRjAAAAHHRFWHRTb2Z0
            d2FyZQBBZG9iZSBGaXJld29ya3MgQ1M1cbXjNgAAAZ1JREFUWIXtl7FKA0EQhr+TgIFgo5BXyBUp
            fIGksLawUNAXWFFfwCJgBAtfIJFMLXgQn8BSwdpCiPcKAdOIoI2x2Dmyd7kYwXhp9odluX/uZv6d
            nZu7DXowxiKZi0IAUHKCvxcsoAIEpST4IawVGb0Hb0BlpcigefACvAAvwAsoTTGGlwwzBAyivLUP
            EZrOM10AhGOH2wWugVVlHoAdhJHrPC8DNR0JGsAAQ9mxNzBOMNjS4Qrq69U5EKmf12ywWVsQI4QI
            IbCn3Gnmnk7uk1bokfooI7QRDlQIGCdzPwiYh0idtXNs2zq3UqwVEiDcu/R0DVjUnFpItuPSscfA
            FXCGSfEAdZ2fVeQ68OjYWwi3ycVvMhABGwgfKXZScHeZ+4c6VzN8FbuYukvOykCs+z8PJ0xqIXYE
            d4ALoKlVH2IIgUHWwd/6gNAFPjPcCPvKNTDcYAj1lXzKc7GIRrSZI6yJzcQ+dtV9bD+IkHThBj34
            4j9/yYxupaQbXPJLNqsGFgeZ6qwpLP1b4AV4AV5AoKfjpR5OwR6VKwULCAC+AQV4W9Ps4uZQAAAA
            AElFTkSuQmCC
            --this_is_a_boundary--
            """
        )

        msg = AnymailInboundMessage.parse_raw_mime(raw)
        attachment = msg.attachments[0]
        attachment_file = attachment.as_uploaded_file()

        self.assertEqual(attachment_file.name, "sample_image.png")
        self.assertEqual(attachment_file.content_type, "image/png")
        self.assertEqual(attachment_file.read(), SAMPLE_IMAGE_CONTENT)

    def test_attachment_as_uploaded_file_security(self):
        # Raw attachment filenames can be malicious; we want to make sure that
        # our Django file converter sanitizes them (as much as any uploaded filename)
        raw = dedent(
            """\
            MIME-Version: 1.0
            Subject: Attachment test
            Content-Type: multipart/mixed; boundary="this_is_a_boundary"

            --this_is_a_boundary
            Content-Type: text/plain; charset="UTF-8"

            The malicious attachment filenames below need to get sanitized

            --this_is_a_boundary
            Content-Type: text/plain; name="report.txt"
            Content-Disposition: attachment; filename="/etc/passwd"

            # (not that overwriting /etc/passwd is actually a thing
            # anymore, but you get the point)
            --this_is_a_boundary
            Content-Type: text/html
            Content-Disposition: attachment; filename="../static/index.html"

            <body>Hey, did I overwrite your site?</body>
            --this_is_a_boundary--
            """
        )
        msg = AnymailInboundMessage.parse_raw_mime(raw)
        attachments = msg.attachments

        # you wouldn't want to actually write here:
        self.assertEqual(attachments[0].get_filename(), "/etc/passwd")
        # path removed - good!:
        self.assertEqual(attachments[0].as_uploaded_file().name, "passwd")

        # ditto for relative paths:
        self.assertEqual(attachments[1].get_filename(), "../static/index.html")
        self.assertEqual(attachments[1].as_uploaded_file().name, "index.html")


class AnymailInboundMessageAttachedMessageTests(SimpleTestCase):
    # message/rfc822 attachments should get parsed recursively

    original_raw_message = dedent(
        """\
        MIME-Version: 1.0
        From: sender@example.com
        Subject: Original message
        Return-Path: bounces@inbound.example.com
        Content-Type: multipart/related; boundary="boundary-orig"

        --boundary-orig
        Content-Type: text/html; charset="UTF-8"

        <img src="cid:abc123"> Here is your message!

        --boundary-orig
        Content-Type: image/png; name="sample_image.png"
        Content-Disposition: inline
        Content-ID: <abc123>
        Content-Transfer-Encoding: base64

        {image_content_base64}
        --boundary-orig--
        """
    ).format(image_content_base64=b64encode(SAMPLE_IMAGE_CONTENT).decode("ascii"))

    def test_parse_rfc822_attachment_from_raw_mime(self):
        # message/rfc822 attachments should be parsed recursively
        raw = dedent(
            """\
            MIME-Version: 1.0
            From: mailer-demon@example.org
            Subject: Undeliverable
            To: bounces@inbound.example.com
            Content-Type: multipart/mixed; boundary="boundary-bounce"

            --boundary-bounce
            Content-Type: text/plain

            Your message was undeliverable due to carrier pigeon strike.
            The original message is attached.

            --boundary-bounce
            Content-Type: message/rfc822
            Content-Disposition: attachment

            {original_raw_message}
            --boundary-bounce--
            """
        ).format(original_raw_message=self.original_raw_message)

        msg = AnymailInboundMessage.parse_raw_mime(raw)
        self.assertIsInstance(msg, AnymailInboundMessage)

        att = msg.get_payload(1)
        self.assertIsInstance(att, AnymailInboundMessage)
        self.assertEqual(att.get_content_type(), "message/rfc822")
        self.assertTrue(att.is_attachment())

        orig_msg = att.get_payload(0)
        self.assertIsInstance(orig_msg, AnymailInboundMessage)
        self.assertEqual(orig_msg["Subject"], "Original message")
        self.assertEqual(orig_msg.get_content_type(), "multipart/related")
        self.assertEqual(att.get_content_text(), self.original_raw_message)

        orig_inline_att = orig_msg.get_payload(1)
        self.assertEqual(orig_inline_att.get_content_type(), "image/png")
        self.assertTrue(orig_inline_att.is_inline())
        self.assertEqual(orig_inline_att.get_payload(decode=True), SAMPLE_IMAGE_CONTENT)

        with self.assertWarnsMessage(
            AnymailDeprecationWarning,
            "is_inline_attachment has been renamed to is_inline and will be removed in "
            "the near future.",
        ):
            self.assertTrue(orig_inline_att.is_inline_attachment())

    def test_construct_rfc822_attachment_from_data(self):
        # constructed message/rfc822 attachment should end up as parsed message
        # (same as if attachment was parsed from raw mime, as in previous test)
        att = AnymailInboundMessage.construct_attachment(
            "message/rfc822", self.original_raw_message
        )
        self.assertIsInstance(att, AnymailInboundMessage)
        self.assertEqual(att.get_content_type(), "message/rfc822")
        self.assertTrue(att.is_attachment())
        self.assertEqual(att.get_content_text(), self.original_raw_message)

        orig_msg = att.get_payload(0)
        self.assertIsInstance(orig_msg, AnymailInboundMessage)
        self.assertEqual(orig_msg["Subject"], "Original message")
        self.assertEqual(orig_msg.get_content_type(), "multipart/related")


class EmailParserBehaviorTests(SimpleTestCase):
    # Python 3.5+'s EmailParser should handle all of these, so long as it's not
    # invoked with its default policy=compat32. This double checks we're using it
    # properly. (Also, older versions of Anymail included workarounds for these
    # in older, broken versions of the EmailParser.)

    def test_parse_folded_headers(self):
        raw = dedent(
            """\
            Content-Type: text/plain
            Subject: This subject uses
             header folding
            X-Json: {"problematic":
             ["encoded newline\\n",
             "comma,semi;no space"]}

            Not-A-Header: This is the body.
             It is not folded.
            """
        )
        for end in ("\n", "\r", "\r\n"):  # check NL, CR, and CRNL line-endings
            msg = AnymailInboundMessage.parse_raw_mime(raw.replace("\n", end))
            self.assertEqual(msg["Subject"], "This subject uses header folding")
            self.assertEqual(
                msg["X-Json"],
                '{"problematic": ["encoded newline\\n", "comma,semi;no space"]}',
            )
            self.assertEqual(
                msg.get_content_text(),
                "Not-A-Header: This is the body.{end} It is not folded.{end}".format(
                    end=end
                ),
            )
            self.assertEqual(msg.defects, [])

    def test_parse_encoded_headers(self):
        # RFC2047 header encoding
        raw = dedent(
            """\
            Content-Type: text/plain
            From: =?US-ASCII?Q?Keith_Moore?= <moore@example.com>
            To: =?ISO-8859-1?Q?Keld_J=F8rn_Simonsen?= <keld@example.com>,
             =?ISO-8859-1?Q?Andr=E9?= "Pirard, Jr." <PIRARD@example.com>
            Cc: =?utf-8?b?TmfGsOG7nWkgbmjhuq1u?= <cc@xn--th-e0a.example.com>
            Subject: =?ISO-8859-1?B?SWYgeW91IGNhbiByZWFkIHRoaXMgeW8=?=
             =?utf-8?q?u_understand_the_example=E2=9C=93?=
            X-Broken: =?utf-8?q?Not_a_char:_=88.?=

            Some examples adapted from http://dogmamix.com/MimeHeadersDecoder/
            """
        )
        msg = AnymailInboundMessage.parse_raw_mime(raw)

        self.assertEqual(msg["From"], "Keith Moore <moore@example.com>")
        self.assertEqual(msg.from_email.display_name, "Keith Moore")
        self.assertEqual(msg.from_email.addr_spec, "moore@example.com")

        self.assertEqual(
            msg["To"],
            "Keld Jørn Simonsen <keld@example.com>, "
            '"André Pirard, Jr." <PIRARD@example.com>',
        )
        self.assertEqual(msg.to[0].display_name, "Keld Jørn Simonsen")
        self.assertEqual(msg.to[1].display_name, "André Pirard, Jr.")

        # Note: Like email.headerregistry.Address, Anymail decodes an RFC2047-encoded
        # display_name, but does not decode a punycode domain.
        # (Use `idna.decode(domain)` if you need that.)
        self.assertEqual(msg["Cc"], "Người nhận <cc@xn--th-e0a.example.com>")
        self.assertEqual(msg.cc[0].display_name, "Người nhận")
        self.assertEqual(msg.cc[0].addr_spec, "cc@xn--th-e0a.example.com")
        self.assertEqual(msg.cc[0].domain, "xn--th-e0a.example.com")

        # Subject breaks between 'o' and 'u' in the word "you", must be re-joined
        # without space. Also tests joining encoded words with different charsets:
        self.assertEqual(
            msg["Subject"],
            "If you can read this you understand the example\N{CHECK MARK}",
        )

        # Replace illegal encodings (rather than causing error):
        self.assertEqual(msg["X-Broken"], "Not a char: \N{REPLACEMENT CHARACTER}.")

    def test_parse_encoded_params(self):
        raw = dedent(
            """\
            MIME-Version: 1.0
            Content-Type: multipart/mixed; boundary="this_is_a_boundary"

            --this_is_a_boundary
            Content-Type: text/plain; charset="UTF-8"

            This is the body

            --this_is_a_boundary
            Content-Type: text/plain; name*=us-ascii''TPS%20Report
            Content-Disposition: attachment;
             filename*=iso-8859-1''Une%20pi%E8ce%20jointe%2Etxt

            This is an attachment
            --this_is_a_boundary--
            """
        )
        msg = AnymailInboundMessage.parse_raw_mime(raw)
        att = msg.attachments[0]
        self.assertTrue(att.is_attachment())
        self.assertEqual(att.get_content_disposition(), "attachment")
        self.assertEqual(
            collapse_rfc2231_value(att.get_param("Name", header="Content-Type")),
            "TPS Report",
        )
        self.assertEqual(att.get_filename(), "Une pièce jointe.txt")



from unittest.mock import ANY

import responses
from django.test import override_settings, tag
from responses.matchers import header_matcher

from anymail.exceptions import AnymailConfigurationError
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.brevo import BrevoInboundWebhookView

from .utils import sample_email_content, sample_image_content
from .webhook_cases import WebhookTestCase


@tag("brevo")
@override_settings(ANYMAIL_BREVO_API_KEY="test-api-key")
class BrevoInboundTestCase(WebhookTestCase):
    def test_inbound_basics(self):
        # Actual (sanitized) Brevo inbound message payload 7/2023
        raw_event = {
            "Uuid": ["aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"],
            "MessageId": "<ABCDE12345@mail.example.com>",
            "InReplyTo": None,
            "From": {"Name": "Sender Name", "Address": "from@example.com"},
            "To": [{"Name": None, "Address": "test@anymail.dev"}],
            "Cc": [{"Name": None, "Address": "test+cc@anymail.dev"}],
            "ReplyTo": None,
            "SentAtDate": "Mon, 17 Jul 2023 11:11:22 -0700",
            "Subject": "Testing Brevo inbound",
            "Attachments": [],
            "Headers": {
                # Headers that appear more than once arrive as lists:
                "Received": [
                    "by outbound.example.com for <test@anymail.dev>; ...",
                    "from [10.10.1.22] by smtp.example.com for <test@anymail.dev>; ...",
                ],
                # Single appearance headers arrive as strings:
                "DKIM-Signature": "v=1; a=rsa-sha256; d=example.com; ...",
                "MIME-Version": "1.0",
                "From": "Sender Name <from@example.com>",
                "Date": "Mon, 17 Jul 2023 11:11:22 -0700",
                "Message-ID": "<ABCDE12345@mail.example.com>",
                "Subject": "Testing Brevo inbound",
                "To": "test@anymail.dev",
                "Cc": "test+cc@anymail.dev",
                "Content-Type": "multipart/alternative",
            },
            "SpamScore": 2.9,
            "ExtractedMarkdownMessage": "This is a test message.  \n",
            "ExtractedMarkdownSignature": "- Sender  \n",
            "RawHtmlBody": '<div dir="ltr">This is a <u>test message</u>.<div><br></div><div>- Mike</div><div><br></div></div>\r\n',  # NOQA: E501
            "RawTextBody": "This is a *test message*.\r\n\n- Sender\r\n",
        }

        response = self.client.post(
            "/anymail/brevo/inbound/",
            content_type="application/json",
            data={"items": [raw_event]},
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=BrevoInboundWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        # Brevo doesn't provide inbound event timestamp
        self.assertIsNone(event.timestamp)
        self.assertEqual(event.event_id, "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Sender Name")
        self.assertEqual(message.from_email.addr_spec, "from@example.com")
        self.assertEqual(
            [str(e) for e in message.to],
            ["test@anymail.dev"],
        )
        self.assertEqual([str(e) for e in message.cc], ["test+cc@anymail.dev"])
        self.assertEqual(message.subject, "Testing Brevo inbound")
        self.assertEqual(message.date.isoformat(" "), "2023-07-17 11:11:22-07:00")
        self.assertEqual(message.text, "This is a *test message*.\r\n\n- Sender\r\n")
        self.assertEqual(
            message.html,
            '<div dir="ltr">This is a <u>test message</u>.<div><br></div><div>- Mike</div><div><br></div></div>\r\n',  # NOQA: E501
        )

        self.assertIsNone(message.envelope_sender)
        self.assertIsNone(message.envelope_recipient)

        # Treat Brevo's ExtractedMarkdownMessage as stripped_text:
        self.assertEqual(message.stripped_text, "This is a test message.  \n")
        # Brevo doesn't provide stripped html:
        self.assertIsNone(message.stripped_html)
        self.assertIsNone(message.spam_detected)
        self.assertEqual(message.spam_score, 2.9)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<ABCDE12345@mail.example.com>")
        self.assertEqual(
            message.get_all("Received"),
            [
                "by outbound.example.com for <test@anymail.dev>; ...",
                "from [10.10.1.22] by smtp.example.com for <test@anymail.dev>; ...",
            ],
        )

    def test_envelope_attrs(self):
        # Brevo's example payload shows Return-Path and Delivered-To headers.
        # They don't seem to be present in our tests, but handle them if they're there:
        raw_event = {
            "Headers": {
                "Return-Path": "<sender@example.com>",
                "Delivered-To": "recipient@example.org",
            }
        }
        self.client.post(
            "/anymail/brevo/inbound/",
            content_type="application/json",
            data={"items": [raw_event]},
        )
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=BrevoInboundWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        message = event.message
        self.assertEqual(message.envelope_sender, "sender@example.com")
        self.assertEqual(message.envelope_recipient, "recipient@example.org")

    @responses.activate
    def test_attachments(self):
        text_content = "Une pièce jointe"
        image_content = sample_image_content()
        email_content = sample_email_content()

        raw_event = {
            # ... much of payload omitted ...
            "Headers": {
                "Content-Type": "multipart/mixed",
            },
            "Attachments": [
                {
                    "Name": "test.txt",
                    "ContentType": "text/plain",
                    "ContentLength": len(text_content),
                    "ContentID": None,
                    "DownloadToken": "download-token-text",
                },
                {
                    "Name": "image.png",
                    "ContentType": "image/png",
                    "ContentLength": len(image_content),
                    "ContentID": "abc123",
                    "DownloadToken": "download-token-image",
                },
                {
                    "Name": "",
                    "ContentType": "message/rfc822",
                    "ContentLength": len(email_content),
                    "ContentID": None,
                    "DownloadToken": "download-token-email",
                },
            ],
        }

        # Brevo supplies a "DownloadToken" that must be used to fetch
        # attachment content. Mock those fetches:
        match_api_key = header_matcher({"api-key": "test-api-key"})
        responses.add(
            responses.GET,
            "https://api.brevo.com/v3/inbound/attachments/download-token-text",
            match=[match_api_key],
            content_type="text/plain; charset=iso-8859-1",
            headers={"content-disposition": 'attachment; filename="test.txt"'},
            body=text_content.encode("iso-8859-1"),
        )
        responses.add(
            responses.GET,
            "https://api.brevo.com/v3/inbound/attachments/download-token-image",
            match=[match_api_key],
            content_type="image/png",
            headers={"content-disposition": 'attachment; filename="image.png"'},
            body=image_content,
        )
        responses.add(
            responses.GET,
            "https://api.brevo.com/v3/inbound/attachments/download-token-email",
            match=[match_api_key],
            content_type="message/rfc822; charset=us-ascii",
            headers={"content-disposition": "attachment"},
            body=email_content,
        )

        response = self.client.post(
            "/anymail/brevo/inbound/",
            content_type="application/json",
            data={"items": [raw_event]},
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=BrevoInboundWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        message = event.message
        attachments = message.attachments  # AnymailInboundMessage convenience accessor
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].get_filename(), "test.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "Une pièce jointe")
        self.assertEqual(attachments[1].get_content_type(), "message/rfc822")
        self.assertEqualIgnoringHeaderFolding(
            attachments[1].get_content_bytes(), email_content
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["abc123"]
        self.assertEqual(inline.get_filename(), "image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), image_content)

    def test_misconfigured_tracking(self):
        errmsg = (
            "You seem to have set Brevo's *tracking* webhook URL"
            " to Anymail's Brevo *inbound* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/brevo/inbound/",
                content_type="application/json",
                data={"event": "delivered"},
            )



import hashlib
import hmac
import json
from datetime import datetime, timezone
from unittest.mock import ANY

from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings, tag

from anymail.exceptions import AnymailConfigurationError
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.mailersend import MailerSendTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase

TEST_WEBHOOK_SIGNING_SECRET = "TEST_WEBHOOK_SIGNING_SECRET"


def mailersend_signature(data, secret):
    """Generate a MailerSend webhook signature for data with secret"""
    # https://developers.mailersend.com/api/v1/webhooks.html#security
    return hmac.new(
        key=secret.encode("ascii"),
        msg=data,
        digestmod=hashlib.sha256,
    ).hexdigest()


class MailerSendWebhookTestCase(WebhookTestCase):
    def client_post_signed(self, url, json_data, secret=TEST_WEBHOOK_SIGNING_SECRET):
        """Return self.client.post(url, serialized json_data) signed with secret"""
        # MailerSend for some reason backslash-escapes all forward slashes ("/")
        # in its webhook payloads ("https:\/\/www..."). This is unnecessary, but
        # harmless. We emulate it here to make sure it won't cause problems.
        data = json.dumps(json_data).replace("/", "\\/").encode("ascii")
        signature = mailersend_signature(data, secret)
        return self.client.post(
            url, content_type="application/json", data=data, HTTP_SIGNATURE=signature
        )


@tag("mailersend")
class MailerSendWebhookSettingsTestCase(MailerSendWebhookTestCase):
    def test_requires_signing_secret(self):
        with self.assertRaisesMessage(
            ImproperlyConfigured, "MAILERSEND_SIGNING_SECRET"
        ):
            self.client_post_signed(
                "/anymail/mailersend/tracking/", {"data": {"type": "sent"}}
            )

    @override_settings(
        ANYMAIL={
            "MAILERSEND_SIGNING_SECRET": "webhook secret",
            "MAILERSEND_INBOUND_SECRET": "inbound secret",
        }
    )
    def test_inbound_secret_is_different(self):
        response = self.client_post_signed(
            "/anymail/mailersend/tracking/",
            {"data": {"type": "sent"}},
            secret="webhook secret",
        )
        self.assertEqual(response.status_code, 200)

    @override_settings(ANYMAIL_MAILERSEND_SIGNING_SECRET="settings secret")
    def test_signing_secret_view_params(self):
        """Webhook signing secret can be provided as a view param"""
        view = MailerSendTrackingWebhookView.as_view(signing_secret="view-level secret")
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(view_instance.signing_secret, b"view-level secret")


@tag("mailersend")
@override_settings(ANYMAIL_MAILERSEND_SIGNING_SECRET=TEST_WEBHOOK_SIGNING_SECRET)
class MailerSendWebhookSecurityTestCase(
    MailerSendWebhookTestCase, WebhookBasicAuthTestCase
):
    should_warn_if_no_auth = False  # because we check webhook signature

    def call_webhook(self):
        return self.client_post_signed(
            "/anymail/mailersend/tracking/",
            {"data": {"type": "sent"}},
            secret=TEST_WEBHOOK_SIGNING_SECRET,
        )

    # Additional tests are in WebhookBasicAuthTestCase

    def test_verifies_correct_signature(self):
        response = self.client_post_signed(
            "/anymail/mailersend/tracking/",
            {"data": {"type": "sent"}},
            secret=TEST_WEBHOOK_SIGNING_SECRET,
        )
        self.assertEqual(response.status_code, 200)

    def test_verifies_missing_signature(self):
        response = self.client.post(
            "/anymail/mailersend/tracking/",
            content_type="application/json",
            data=json.dumps({"data": {"type": "sent"}}),
        )
        self.assertEqual(response.status_code, 400)

    def test_verifies_bad_signature(self):
        # This also verifies that the error log references the correct setting to check.
        with self.assertLogs() as logs:
            response = self.client_post_signed(
                "/anymail/mailersend/tracking/",
                {"data": {"type": "sent"}},
                secret="wrong signing key",
            )
        # SuspiciousOperation causes 400 response (even in test client):
        self.assertEqual(response.status_code, 400)
        self.assertIn("check Anymail MAILERSEND_SIGNING_SECRET", logs.output[0])


@tag("mailersend")
@override_settings(ANYMAIL_MAILERSEND_SIGNING_SECRET=TEST_WEBHOOK_SIGNING_SECRET)
class MailerSendTestCase(MailerSendWebhookTestCase):
    def test_sent_event(self):
        # This is an actual, complete (sanitized) "sent" event as received from
        # MailerSend. (For brevity, later tests omit several payload fields that
        # Anymail doesn't use.)
        raw_event = {
            "type": "activity.sent",
            "domain_id": "[domain-id-redacted]",
            "created_at": "2023-02-27T21:09:49.520507Z",
            "webhook_id": "[webhook-id-redacted]",
            "url": "https://test.anymail.dev/anymail/mailersend/tracking/",
            "data": {
                "object": "activity",
                "id": "63fd1c1d31b9c750540fe85c",
                "type": "sent",
                "created_at": "2023-02-27T21:09:49.506000Z",
                "email": {
                    "object": "email",
                    "id": "63fd1c1de225707fa905f0a8",
                    "created_at": "2023-02-27T21:09:49.141000Z",
                    "from": "sender@mailersend.anymail.dev",
                    "subject": "Test webhooks",
                    "status": "sent",
                    "tags": ["tag1", "Tag 2"],
                    "message": {
                        "object": "message",
                        "id": "63fd1c1d5f010335ed07066b",
                        "created_at": "2023-02-27T21:09:49.061000Z",
                    },
                    "recipient": {
                        "object": "recipient",
                        "id": "63f3bb1965d98aa98c07d6b7",
                        "email": "recipient@example.com",
                        "created_at": "2023-02-20T18:25:29.162000Z",
                    },
                },
                "morph": None,
                "template_id": "",
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "sent")
        # event.timestamp comes from data.created_at:
        self.assertEqual(
            event.timestamp,
            # "2023-02-27T21:09:49.506000Z"
            datetime(2023, 2, 27, 21, 9, 49, microsecond=506000, tzinfo=timezone.utc),
        )
        # event.message_id matches the message.anymail_status.message_id when the
        # message was sent. It comes from data.email.message.id:
        self.assertEqual(event.message_id, "63fd1c1d5f010335ed07066b")
        # event.event_id is unique for each event, and comes from data.id:
        self.assertEqual(event.event_id, "63fd1c1d31b9c750540fe85c")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["tag1", "Tag 2"])
        self.assertEqual(event.metadata, {})  # MailerSend doesn't support metadata
        self.assertEqual(event.esp_event, raw_event)

        # You can construct the sent Message-ID header (which is different from the
        # event.message_id, and is unique per recipient) from esp_event.data.email.id:
        sent_message_id = f"<{event.esp_event['data']['email']['id']}@mailersend.net>"
        self.assertEqual(sent_message_id, "<63fd1c1de225707fa905f0a8@mailersend.net>")

    def test_delivered_event(self):
        raw_event = {
            "type": "activity.delivered",
            "data": {
                "object": "activity",
                "id": "63fd1c1fcfbe46145d003a7b",
                "type": "delivered",
                "created_at": "2023-02-27T21:09:51.865000Z",
                "email": {
                    "status": "delivered",
                    "message": {
                        "id": "63fd1c1d5f010335ed07066b",
                    },
                    "recipient": {
                        "email": "recipient@example.com",
                    },
                },
                "morph": None,
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(
            event.timestamp,
            # "2023-02-27T21:09:51.865000Z"
            datetime(2023, 2, 27, 21, 9, 51, microsecond=865000, tzinfo=timezone.utc),
        )
        self.assertEqual(event.message_id, "63fd1c1d5f010335ed07066b")
        self.assertEqual(event.event_id, "63fd1c1fcfbe46145d003a7b")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_hard_bounced_event(self):
        raw_event = {
            "type": "activity.hard_bounced",
            "data": {
                "id": "63fd251d5c00f8e52001fce6",
                "type": "hard_bounced",
                "created_at": "2023-02-27T21:48:13.593000Z",
                "email": {
                    "status": "rejected",
                    "message": {
                        "id": "63fd25194d5edba3da09e044",
                    },
                    "recipient": {
                        "email": "invalid@example.com",
                    },
                },
                "morph": {
                    "object": "recipient_bounce",
                    "reason": "Host or domain name not found",
                },
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.recipient, "invalid@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "Host or domain name not found")
        self.assertIsNone(event.mta_response)  # raw MTA info not provided

    def test_soft_bounced_event(self):
        raw_event = {
            "type": "activity.soft_bounced",
            "data": {
                "object": "activity",
                "id": "62f114f8165fe0d8db0288e5",
                "type": "soft_bounced",
                "created_at": "2022-08-08T13:51:52.747000Z",
                "email": {
                    "status": "rejected",
                    "tags": None,
                    "message": {
                        "id": "62fb66bef54a112e920b5493",
                    },
                    "recipient": {
                        "email": "notauser@example.com",
                    },
                },
                "morph": {
                    "object": "recipient_bounce",
                    "reason": "Unknown reason",
                },
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.recipient, "notauser@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "Unknown reason")
        self.assertIsNone(event.mta_response)  # raw MTA info not provided

    def test_spam_complaint_event(self):
        raw_event = {
            "type": "activity.spam_complaint",
            "data": {
                "id": "62f114f8165fe0d8db0288e5",
                "type": "spam_complaint",
                "created_at": "2022-08-08T13:51:52.747000Z",
                "email": {
                    "status": "delivered",
                    "message": {
                        "id": "62fb66bef54a112e920b5493",
                    },
                    "recipient": {
                        "email": "recipient@example.com",
                    },
                },
                "morph": {
                    "object": "spam_complaint",
                    "reason": None,
                },
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_unsubscribed_event(self):
        raw_event = {
            "type": "activity.unsubscribed",
            "data": {
                "id": "63fd21c23f2bdd360e07d6b2",
                "type": "unsubscribed",
                "created_at": "2023-02-27T21:33:54.791000Z",
                "email": {
                    "status": "delivered",
                    "message": {
                        "id": "63fd1c1d5f010335ed07066b",
                    },
                    "recipient": {
                        "email": "recipient@example.com",
                    },
                },
                "morph": {
                    "object": "recipient_unsubscribe",
                    "reason": "option_3",
                    "readable_reason": "I get too many emails",
                },
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "unsubscribed")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.description, "I get too many emails")

    def test_opened_event(self):
        raw_event = {
            "type": "activity.opened",
            "data": {
                "id": "63fd1e05532bd6cc700a793b",
                "type": "opened",
                "created_at": "2023-02-27T21:17:57.025000Z",
                "email": {
                    "status": "delivered",
                    "message": {
                        "id": "63fd1c1d5f010335ed07066b",
                    },
                    "recipient": {
                        "email": "recipient@example.com",
                    },
                },
                "morph": {
                    "object": "open",
                    "id": "63fd1e05532bd6cc700a793a",
                    "created_at": "2023-02-27T21:17:57.018000Z",
                    "ip": "10.10.10.10",
                },
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_clicked_event(self):
        raw_event = {
            "type": "activity.clicked",
            "data": {
                "id": "63fd1d23afa3c770b00da7d3",
                "type": "clicked",
                "created_at": "2023-02-27T21:14:11.691000Z",
                "email": {
                    "status": "delivered",
                    "message": {
                        "id": "63fd1c1d5f010335ed07066b",
                    },
                    "recipient": {
                        "email": "recipient@example.com",
                    },
                },
                "morph": {
                    "object": "click",
                    "id": "63fd1d23afa3c770b00da7d2",
                    "created_at": "2023-02-27T21:14:11.679000Z",
                    "ip": "10.10.10.10",
                    "url": "https://example.com/test",
                },
            },
        }
        response = self.client_post_signed("/anymail/mailersend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailerSendTrackingWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.click_url, "https://example.com/test")

    def test_misconfigured_inbound(self):
        errmsg = (
            "You seem to have set MailerSend's *inbound* route endpoint"
            " to Anymail's MailerSend *activity tracking* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client_post_signed(
                "/anymail/mailersend/tracking/",
                {
                    "type": "inbound.message",
                    "data": {"object": "message", "raw": "..."},
                },
            )



from django.urls import path, re_path

from .webhooks.amazon_ses import (
    AmazonSESInboundWebhookView,
    AmazonSESTrackingWebhookView,
)
from .webhooks.brevo import BrevoInboundWebhookView, BrevoTrackingWebhookView
from .webhooks.mailersend import (
    MailerSendInboundWebhookView,
    MailerSendTrackingWebhookView,
)
from .webhooks.mailgun import MailgunInboundWebhookView, MailgunTrackingWebhookView
from .webhooks.mailjet import MailjetInboundWebhookView, MailjetTrackingWebhookView
from .webhooks.mandrill import MandrillCombinedWebhookView
from .webhooks.postal import PostalInboundWebhookView, PostalTrackingWebhookView
from .webhooks.postmark import PostmarkInboundWebhookView, PostmarkTrackingWebhookView
from .webhooks.resend import ResendTrackingWebhookView
from .webhooks.sendgrid import SendGridInboundWebhookView, SendGridTrackingWebhookView
from .webhooks.sendinblue import (
    SendinBlueInboundWebhookView,
    SendinBlueTrackingWebhookView,
)
from .webhooks.sparkpost import (
    SparkPostInboundWebhookView,
    SparkPostTrackingWebhookView,
)
from .webhooks.unisender_go import UnisenderGoTrackingWebhookView

app_name = "anymail"
urlpatterns = [
    path(
        "amazon_ses/inbound/",
        AmazonSESInboundWebhookView.as_view(),
        name="amazon_ses_inbound_webhook",
    ),
    path(
        "brevo/inbound/",
        BrevoInboundWebhookView.as_view(),
        name="brevo_inbound_webhook",
    ),
    path(
        "mailersend/inbound/",
        MailerSendInboundWebhookView.as_view(),
        name="mailersend_inbound_webhook",
    ),
    re_path(
        # Mailgun delivers inbound messages differently based on whether
        # the webhook url contains "mime" (anywhere). You can use either
        # ".../mailgun/inbound/" or ".../mailgun/inbound_mime/" depending
        # on the behavior you want.
        r"^mailgun/inbound(_mime)?/$",
        MailgunInboundWebhookView.as_view(),
        name="mailgun_inbound_webhook",
    ),
    path(
        "mailjet/inbound/",
        MailjetInboundWebhookView.as_view(),
        name="mailjet_inbound_webhook",
    ),
    path(
        "postal/inbound/",
        PostalInboundWebhookView.as_view(),
        name="postal_inbound_webhook",
    ),
    path(
        "postmark/inbound/",
        PostmarkInboundWebhookView.as_view(),
        name="postmark_inbound_webhook",
    ),
    path(
        "sendgrid/inbound/",
        SendGridInboundWebhookView.as_view(),
        name="sendgrid_inbound_webhook",
    ),
    path(
        # Compatibility for old SendinBlue esp_name; use Brevo in new code
        "sendinblue/inbound/",
        SendinBlueInboundWebhookView.as_view(),
        name="sendinblue_inbound_webhook",
    ),
    path(
        "sparkpost/inbound/",
        SparkPostInboundWebhookView.as_view(),
        name="sparkpost_inbound_webhook",
    ),
    path(
        "amazon_ses/tracking/",
        AmazonSESTrackingWebhookView.as_view(),
        name="amazon_ses_tracking_webhook",
    ),
    path(
        "brevo/tracking/",
        BrevoTrackingWebhookView.as_view(),
        name="brevo_tracking_webhook",
    ),
    path(
        "mailersend/tracking/",
        MailerSendTrackingWebhookView.as_view(),
        name="mailersend_tracking_webhook",
    ),
    path(
        "mailgun/tracking/",
        MailgunTrackingWebhookView.as_view(),
        name="mailgun_tracking_webhook",
    ),
    path(
        "mailjet/tracking/",
        MailjetTrackingWebhookView.as_view(),
        name="mailjet_tracking_webhook",
    ),
    path(
        "postal/tracking/",
        PostalTrackingWebhookView.as_view(),
        name="postal_tracking_webhook",
    ),
    path(
        "postmark/tracking/",
        PostmarkTrackingWebhookView.as_view(),
        name="postmark_tracking_webhook",
    ),
    path(
        "resend/tracking/",
        ResendTrackingWebhookView.as_view(),
        name="resend_tracking_webhook",
    ),
    path(
        "sendgrid/tracking/",
        SendGridTrackingWebhookView.as_view(),
        name="sendgrid_tracking_webhook",
    ),
    path(
        # Compatibility for old SendinBlue esp_name; use Brevo in new code
        "sendinblue/tracking/",
        SendinBlueTrackingWebhookView.as_view(),
        name="sendinblue_tracking_webhook",
    ),
    path(
        "sparkpost/tracking/",
        SparkPostTrackingWebhookView.as_view(),
        name="sparkpost_tracking_webhook",
    ),
    path(
        "unisender_go/tracking/",
        UnisenderGoTrackingWebhookView.as_view(),
        name="unisender_go_tracking_webhook",
    ),
    # Anymail uses a combined Mandrill webhook endpoint,
    # to simplify Mandrill's key-validation scheme:
    path("mandrill/", MandrillCombinedWebhookView.as_view(), name="mandrill_webhook"),
    # This url is maintained for backwards compatibility with earlier Anymail releases:
    path(
        "mandrill/tracking/",
        MandrillCombinedWebhookView.as_view(),
        name="mandrill_tracking_webhook",
    ),
]



from __future__ import annotations

from email.headerregistry import Address

from django.test import SimpleTestCase, override_settings, tag
from requests.structures import CaseInsensitiveDict

from anymail.backends.unisender_go import EmailBackend, UnisenderGoPayload
from anymail.message import AnymailMessage

TEMPLATE_ID = "template_id"
FROM_EMAIL = "sender@test.test"
FROM_NAME = "test name"
TO_EMAIL = "receiver@test.test"
TO_NAME = "receiver"
OTHER_TO_EMAIL = "receiver1@test.test"
OTHER_TO_NAME = "receiver1"
SUBJECT = "subject"
GLOBAL_DATA = {"arg": "arg"}
SUBSTITUTION_ONE = {"arg1": "arg1"}
SUBSTITUTION_TWO = {"arg2": "arg2"}


@tag("unisender_go")
@override_settings(ANYMAIL_UNISENDER_GO_API_KEY=None, ANYMAIL_UNISENDER_GO_API_URL="")
class TestUnisenderGoPayload(SimpleTestCase):
    def test_unisender_go_payload__full(self):
        substitutions = {TO_EMAIL: SUBSTITUTION_ONE, OTHER_TO_EMAIL: SUBSTITUTION_TWO}
        email = AnymailMessage(
            template_id=TEMPLATE_ID,
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=str(Address(display_name=FROM_NAME, addr_spec=FROM_EMAIL)),
            to=[
                str(Address(display_name=TO_NAME, addr_spec=TO_EMAIL)),
                str(Address(display_name=OTHER_TO_NAME, addr_spec=OTHER_TO_EMAIL)),
            ],
            merge_data=substitutions,
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {
                "to": ", ".join(email.to),
            },
            "recipients": [
                {
                    "email": TO_EMAIL,
                    "substitutions": {**SUBSTITUTION_ONE, "to_name": TO_NAME},
                },
                {
                    "email": OTHER_TO_EMAIL,
                    "substitutions": {**SUBSTITUTION_TWO, "to_name": OTHER_TO_NAME},
                },
            ],
            "subject": SUBJECT,
            "template_id": TEMPLATE_ID,
        }

        self.assertEqual(payload.data, expected_payload)

    def test_unisender_go_payload__cc_bcc(self):
        cc_to_email = "receiver_cc@test.test"
        bcc_to_email = "receiver_bcc@test.test"
        email = AnymailMessage(
            template_id=TEMPLATE_ID,
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=f"{FROM_NAME} <{FROM_EMAIL}>",
            to=[
                str(Address(display_name=TO_NAME, addr_spec=TO_EMAIL)),
                str(Address(display_name=OTHER_TO_NAME, addr_spec=OTHER_TO_EMAIL)),
            ],
            cc=[cc_to_email],
            bcc=[bcc_to_email],
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_headers = {
            "To": f"{TO_NAME} <{TO_EMAIL}>, {OTHER_TO_NAME} <{OTHER_TO_EMAIL}>",
            "CC": cc_to_email,
        }
        expected_headers = CaseInsensitiveDict(expected_headers)
        expected_recipients = [
            {
                "email": TO_EMAIL,
                "substitutions": {"to_name": TO_NAME},
            },
            {
                "email": OTHER_TO_EMAIL,
                "substitutions": {"to_name": OTHER_TO_NAME},
            },
            {"email": cc_to_email},
            {"email": bcc_to_email},
        ]

        self.assertEqual(payload.data["headers"], expected_headers)
        self.assertCountEqual(payload.data["recipients"], expected_recipients)

    def test_unisender_go_payload__parse_from__with_name(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=str(Address(display_name=FROM_NAME, addr_spec=FROM_EMAIL)),
            to=[TO_EMAIL],
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
        }

        self.assertEqual(payload.data, expected_payload)

    def test_unisender_go_payload__parse_from__without_name(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=FROM_EMAIL,
            to=[TO_EMAIL],
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
        }

        self.assertEqual(payload.data, expected_payload)

    @override_settings(
        ANYMAIL={"UNISENDER_GO_SEND_DEFAULTS": {"esp_extra": {"skip_unsubscribe": 1}}},
    )
    def test_unisender_go_payload__parse_from__with_unsub__in_settings(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=f"{FROM_NAME} <{FROM_EMAIL}>",
            to=[TO_EMAIL],
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
            "skip_unsubscribe": 1,
        }

        self.assertEqual(payload.data, expected_payload)

    @override_settings(
        ANYMAIL={"UNISENDER_GO_SEND_DEFAULTS": {"esp_extra": {"skip_unsubscribe": 0}}},
    )
    def test_unisender_go_payload__parse_from__with_unsub__in_args(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=f"{FROM_NAME} <{FROM_EMAIL}>",
            to=[TO_EMAIL],
            esp_extra={"skip_unsubscribe": 1},
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
            "skip_unsubscribe": 1,
        }

        self.assertEqual(payload.data, expected_payload)

    @override_settings(
        ANYMAIL={
            "UNISENDER_GO_SEND_DEFAULTS": {"esp_extra": {"global_language": "en"}}
        },
    )
    def test_unisender_go_payload__parse_from__global_language__in_settings(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=f"{FROM_NAME} <{FROM_EMAIL}>",
            to=[TO_EMAIL],
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
            "global_language": "en",
        }

        self.assertEqual(payload.data, expected_payload)

    @override_settings(
        ANYMAIL={
            "UNISENDER_GO_SEND_DEFAULTS": {"esp_extra": {"global_language": "fr"}}
        },
    )
    def test_unisender_go_payload__parse_from__global_language__in_args(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=f"{FROM_NAME} <{FROM_EMAIL}>",
            to=[TO_EMAIL],
            esp_extra={"global_language": "en"},
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
            "global_language": "en",
        }

        self.assertEqual(payload.data, expected_payload)

    def test_unisender_go_payload__parse_from__bypass_esp_extra(self):
        email = AnymailMessage(
            subject=SUBJECT,
            merge_global_data=GLOBAL_DATA,
            from_email=f"{FROM_NAME} <{FROM_EMAIL}>",
            to=[TO_EMAIL],
            esp_extra={
                "bypass_global": 1,
                "bypass_unavailable": 1,
                "bypass_unsubscribed": 1,
                "bypass_complained": 1,
            },
        )
        backend = EmailBackend()

        payload = UnisenderGoPayload(
            message=email, backend=backend, defaults=backend.send_defaults
        )
        expected_payload = {
            "from_email": FROM_EMAIL,
            "from_name": FROM_NAME,
            "global_substitutions": GLOBAL_DATA,
            "headers": {"to": TO_EMAIL},
            "recipients": [{"email": TO_EMAIL}],
            "subject": SUBJECT,
            "bypass_global": 1,
            "bypass_unavailable": 1,
            "bypass_unsubscribed": 1,
            "bypass_complained": 1,
        }

        self.assertEqual(payload.data, expected_payload)



import os
import unittest
from datetime import datetime, timedelta
from email.headerregistry import Address

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin

ANYMAIL_TEST_UNISENDER_GO_API_KEY = os.getenv("ANYMAIL_TEST_UNISENDER_GO_API_KEY")
ANYMAIL_TEST_UNISENDER_GO_API_URL = os.getenv("ANYMAIL_TEST_UNISENDER_GO_API_URL")
ANYMAIL_TEST_UNISENDER_GO_DOMAIN = os.getenv("ANYMAIL_TEST_UNISENDER_GO_DOMAIN")
ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID = os.getenv(
    "ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID"
)


@tag("unisender_go", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_UNISENDER_GO_API_KEY
    and ANYMAIL_TEST_UNISENDER_GO_API_URL
    and ANYMAIL_TEST_UNISENDER_GO_DOMAIN,
    "Set ANYMAIL_TEST_UNISENDER_GO_API_KEY, ANYMAIL_TEST_UNISENDER_GO_API_URL"
    " and ANYMAIL_TEST_UNISENDER_GO_DOMAIN environment variables to run Unisender Go"
    " integration tests",
)
@override_settings(
    ANYMAIL_UNISENDER_GO_API_KEY=ANYMAIL_TEST_UNISENDER_GO_API_KEY,
    ANYMAIL_UNISENDER_GO_API_URL=ANYMAIL_TEST_UNISENDER_GO_API_URL,
    EMAIL_BACKEND="anymail.backends.unisender_go.EmailBackend",
)
class UnisenderGoBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """
    Unisender Go API integration tests

    These tests run against the **live** Unisender Go API, using the
    environment variable `ANYMAIL_TEST_UNISENDER_GO_API_KEY` as the API key,
    `ANYMAIL_UNISENDER_GO_API_URL` as the API URL where that key was issued,
    and `ANYMAIL_TEST_UNISENDER_GO_DOMAIN` to construct the sender addresses.
    If any of those variables are not set, these tests won't run.

    To run the template test, also set ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID
    to a valid template in your account.

    The tests send actual email to a sink address at anymail.dev.
    """

    def setUp(self):
        super().setUp()
        self.from_email = f"from@{ANYMAIL_TEST_UNISENDER_GO_DOMAIN}"
        self.message = AnymailMessage(
            "Anymail Unisender Go integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Unisender Go send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "queued")  # Unisender Go always queues
        self.assertRegex(message_id, r".+")
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        send_at = datetime.now() + timedelta(minutes=2)
        message = AnymailMessage(
            subject="Anymail Unisender Go all-options integration test",
            body="This is the text body",
            from_email=str(
                Address(display_name="Test From, with comma", addr_spec=self.from_email)
            ),
            to=["test+to1@anymail.dev", '"Recipient 2, OK?" <test+to2@anymail.dev>'],
            cc=["test+cc1@anymail.dev", '"Copy 2, OK?" <test+cc2@anymail.dev>'],
            bcc=[
                f"test+bcc1@{ANYMAIL_TEST_UNISENDER_GO_DOMAIN}",
                f'"BCC 2, OK?" <test+bcc2@{ANYMAIL_TEST_UNISENDER_GO_DOMAIN}>',
            ],
            # Unisender Go only supports a single reply-to:
            reply_to=['"Reply, with comma (and parens)" <reply@example.com>'],
            headers={"X-Anymail-Test": "value", "X-Anymail-Count": 3},
            metadata={"meta1": "simple string", "meta2": 2},
            send_at=send_at,
            tags=["tag 1", "tag 2"],
            track_opens=False,
            track_clicks=False,
            esp_extra={
                "global_language": "en",
                "options": {"unsubscribe_url": "https://example.com/unsubscribe?id=1"},
            },
        )
        message.attach_alternative("<p>HTML content</p>", "text/html")
        message.attach_alternative("<p>AMP HTML content</p>", "text/x-amp-html")

        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")

        message.send()
        self.assertEqual(message.anymail_status.status, {"queued"})
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")
        self.assertRegex(recipient_status["test+to1@anymail.dev"].message_id, r".+")
        self.assertRegex(recipient_status["test+to2@anymail.dev"].message_id, r".+")
        # Anymail generates unique message_id for each recipient:
        self.assertNotEqual(
            recipient_status["test+to1@anymail.dev"].message_id,
            recipient_status["test+to2@anymail.dev"].message_id,
        )

    @unittest.skipUnless(
        ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID,
        "Set ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID to run the"
        " Unisender Go template integration test",
    )
    def test_template(self):
        """
        To run this test, create a template in your account containing
        "{{order_id}}" and "{{ship_date}}" substitutions, and set
        ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID to the template's id.
        """
        message = AnymailMessage(
            # This is an actual template in the Anymail test account:
            template_id=ANYMAIL_TEST_UNISENDER_GO_TEMPLATE_ID,
            to=["Recipient 1 <test+to1@anymail.dev>", "test+to2@anymail.dev"],
            reply_to=["Do not reply <reply@example.dev>"],
            tags=["using-template"],
            merge_data={
                "test+to1@anymail.dev": {"order_id": "12345"},
                "test+to2@anymail.dev": {"order_id": "23456"},
            },
            merge_global_data={"ship_date": "yesterday"},
            metadata={"customer-id": "unknown", "meta2": 2},
            merge_metadata={
                "test+to1@anymail.dev": {"customer-id": "ZXK9123"},
                "test+to2@anymail.dev": {"customer-id": "ZZT4192"},
            },
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "test+to1@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "test+to2@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.from_email = None  # use template sender
        message.attach("attachment1.txt", "Here is some\ntext", "text/plain")

        message.send()
        # Unisender Go always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")
        self.assertRegex(recipient_status["test+to1@anymail.dev"].message_id, r".+")
        self.assertRegex(recipient_status["test+to2@anymail.dev"].message_id, r".+")
        # Anymail generates unique message_id for each recipient:
        self.assertNotEqual(
            recipient_status["test+to1@anymail.dev"].message_id,
            recipient_status["test+to2@anymail.dev"].message_id,
        )

    @override_settings(ANYMAIL_UNISENDER_GO_API_KEY="Hey, that's not an API key!")
    def test_invalid_api_key(self):
        # Make sure the exception message includes Unisender Go's response:
        with self.assertRaisesMessage(AnymailAPIError, "Can not decode key"):
            self.message.send()
